"""
Module for serialization and deserialization of KNX/IP CEMI Frame.

cEMI stands for Common External Message Interface

A CEMI frame is the container to transport a KNX/IP Telegram to/from the KNX bus.

Documentation within:

    Application Note 117/08 v02
    KNX IP Communication Medium
    File: AN117 v02.01 KNX IP Communication Medium DV.pdf
"""

from __future__ import annotations

from abc import ABC, abstractmethod

from xknx.exceptions import ConversionError, CouldNotParseCEMI, UnsupportedCEMIMessage
from xknx.profile.const import ResourceObjectType, ResourcePropertyId
from xknx.telegram import GroupAddress, IndividualAddress, Telegram
from xknx.telegram.apci import APCI
from xknx.telegram.tpci import TPCI, TDataBroadcast

from .const import CEMIErrorCode, CEMIFlags, CEMIMessageCode


class CEMIInfo:
    """
    Raw helper class for all CEMI additional info.

    As specified within KNX Chapter 3.6.3/4.1.4.3 "Additional information".
    """

    def __init__(self, raw: bytes = bytes(0)) -> None:
        """Initialize CEMILInfo object."""
        self.raw = raw

    def calculated_length(self) -> int:
        """Get length of CEMI info."""
        return 1 + len(self.raw)

    @staticmethod
    def from_knx(raw: bytes) -> tuple[CEMIInfo, bytes]:
        """Parse/deserialize from CEMI raw data."""
        length = raw[0]
        return CEMIInfo(raw[1 : length + 1]), raw[length + 1 :]

    def to_knx(self) -> bytes:
        """Serialize to KNX/IP raw data."""
        return bytes((len(self.raw),)) + self.raw

    def __repr__(self) -> str:
        """Return object as readable string."""
        return f'CEMIInfo("{self.raw.hex()}")'

    def __eq__(self, other: object) -> bool:
        """Equal operator."""
        return self.__dict__ == other.__dict__


class CEMIData(ABC):
    """Base class for all CEMI data."""

    @abstractmethod
    def calculated_length(self) -> int:
        """Get length of CEMI data."""

    @abstractmethod
    def to_knx(self) -> bytes:
        """Serialize to CEMI raw data."""

    @classmethod
    @abstractmethod
    def from_knx(cls, raw: bytes) -> CEMIData:
        """Parse/deserialize from KNX/IP raw data."""

    @abstractmethod
    def __repr__(self) -> str:
        """Return object as readable string."""

    def __eq__(self, other: object) -> bool:
        """Equal operator."""
        if self.__class__ != other.__class__:
            return False
        return self.__dict__ == other.__dict__


class CEMILData(CEMIData):
    """Representation of CEMI Link Layer Data."""

    def __init__(
        self,
        *,
        flags: int,
        src_addr: IndividualAddress,
        dst_addr: GroupAddress | IndividualAddress,
        tpci: TPCI,
        payload: APCI | None,
    ) -> None:
        """Initialize CEMILData object."""
        self.flags = flags
        self.src_addr = src_addr
        self.dst_addr = dst_addr
        self.tpci = tpci
        self.payload = payload

    @property
    def hops(self) -> int:
        """Return hops."""
        return (self.flags & 0x0070) >> 4

    @hops.setter
    def hops(self, val: int) -> None:
        """Set hops."""
        # Resetting hops
        self.flags &= 0xFFFF ^ 0x0070
        # Setting new hops
        self.flags |= val << 4

    def calculated_length(self) -> int:
        """Get length of KNX/IP body."""
        if not self.tpci.control and self.payload is not None:
            return 8 + self.payload.calculated_length()
        if self.tpci.control and self.payload is None:
            return 8
        raise TypeError("Data TPDU must have a payload; control TPDU must not.")

    @staticmethod
    def init_from_telegram(
        telegram: Telegram,
        src_addr: IndividualAddress | None = None,
    ) -> CEMILData:
        """Return CEMILData from a Telegram."""
        flags = (
            CEMIFlags.FRAME_TYPE_STANDARD
            | CEMIFlags.DO_NOT_REPEAT
            | CEMIFlags.BROADCAST
            | CEMIFlags.NO_ACK_REQUESTED
            | CEMIFlags.CONFIRM_NO_ERROR
            | CEMIFlags.HOP_COUNT_1ST
        )
        if isinstance(telegram.destination_address, GroupAddress):
            flags |= CEMIFlags.DESTINATION_GROUP_ADDRESS
            if isinstance(telegram.tpci, TDataBroadcast):
                flags |= CEMIFlags.PRIORITY_SYSTEM
            else:
                flags |= CEMIFlags.PRIORITY_LOW
        elif isinstance(telegram.destination_address, IndividualAddress):
            flags |= (
                CEMIFlags.DESTINATION_INDIVIDUAL_ADDRESS | CEMIFlags.PRIORITY_SYSTEM
            )
        else:
            raise TypeError()

        return CEMILData(
            flags=flags,
            src_addr=src_addr or telegram.source_address,
            dst_addr=telegram.destination_address,
            tpci=telegram.tpci,
            payload=telegram.payload,
        )

    def telegram(self) -> Telegram:
        """Return Telegram from a CEMILData."""

        return Telegram(
            destination_address=self.dst_addr,
            payload=self.payload,
            source_address=self.src_addr,
            tpci=self.tpci,
        )

    def to_knx(self) -> bytes:
        """Serialize to KNX/IP raw data."""
        tpdu: bytes | bytearray
        if self.tpci.control:
            tpdu = self.tpci.to_knx().to_bytes(1, "big")
            npdu_len = 0
        else:
            if not isinstance(self.payload, APCI):
                raise ConversionError(
                    f"Invalid payload set for data TPDU: {type(self.payload)}"
                )
            tpdu = self.payload.to_knx()
            tpdu[0] |= self.tpci.to_knx()
            npdu_len = self.payload.calculated_length()

        return (
            self.flags.to_bytes(2, "big")
            + self.src_addr.to_knx()
            + self.dst_addr.to_knx()
            + npdu_len.to_bytes(1, "big")
            + tpdu
        )

    @classmethod
    def from_knx(cls, raw: bytes) -> CEMILData:
        """Parse L_DATA_IND, CEMIMessageCode.L_DATA_REQ, CEMIMessageCode.L_DATA_CON."""
        if len(raw) < 8:
            raise CouldNotParseCEMI(
                f"CEMI too small. Length: {len(raw)}; CEMI: {raw.hex()}"
            )

        # Control field 1 and Control field 2 - first 2 octets
        flags = int.from_bytes(raw[0:2], "big")

        src_addr = IndividualAddress.from_knx(raw[2:4])

        _dst_is_group_address = bool(flags & CEMIFlags.DESTINATION_GROUP_ADDRESS)
        dst_addr: GroupAddress | IndividualAddress = (
            GroupAddress.from_knx(raw[4:6])
            if _dst_is_group_address
            else IndividualAddress.from_knx(raw[4:6])
        )

        _npdu_len = raw[6]
        _tpdu = raw[7:]
        _apdu = bytes([_tpdu[0] & 0b11]) + _tpdu[1:]  # clear TPCI bits
        if len(_apdu) != (_npdu_len + 1):  # TCPI octet not included in NPDU length
            raise CouldNotParseCEMI(
                f"APDU LEN should be {_npdu_len} but is {len(_apdu) - 1} "
                f"from {src_addr} in CEMI: {raw.hex()}"
            )

        # TPCI (transport layer control information)
        # - with control bit set -> 8 bit; no APDU
        # - no control bit set (data) -> First 6 bit
        # APCI (application layer control information) -> Last  10 bit of TPCI/APCI
        try:
            tpci = TPCI.resolve(
                raw_tpci=_tpdu[0],
                dst_is_group_address=_dst_is_group_address,
                dst_is_zero=not dst_addr.raw,
            )
        except ConversionError as err:
            raise UnsupportedCEMIMessage(
                f"TPCI not supported: {_tpdu[0]:#10b} "
                f"from {src_addr} in CEMI: {raw.hex()}"
            ) from err

        if tpci.control:
            if _npdu_len:
                raise CouldNotParseCEMI(
                    f"Invalid length for control TPDU {tpci}: {_npdu_len}"
                    f" from {src_addr} in CEMI: {raw.hex()}"
                )
            return cls(
                flags=flags,
                src_addr=src_addr,
                dst_addr=dst_addr,
                tpci=tpci,
                payload=None,
            )

        try:
            payload = APCI.from_knx(_apdu)
        except ConversionError as err:
            raise UnsupportedCEMIMessage(
                f"APDU not supported from {src_addr} "
                f"from {src_addr} in CEMI: {raw.hex()}"
            ) from err

        return cls(
            flags=flags,
            src_addr=src_addr,
            dst_addr=dst_addr,
            tpci=tpci,
            payload=payload,
        )

    def __repr__(self) -> str:
        """Return object as readable string."""
        return (
            "CEMILData("
            f'src_addr="{self.src_addr.__repr__()}" '
            f'dst_addr="{self.dst_addr.__repr__()}" '
            f'flags="{self.flags:16b}" '
            f'tpci="{self.tpci}" '
            f'payload="{self.payload}")'
        )


class CEMIMPropInfo:
    """Representation of CEMI Device Management Property."""

    LENGTH = 6

    def __init__(
        self,
        *,
        object_type: ResourceObjectType,
        object_instance: int = 1,
        property_id: ResourcePropertyId | int,
        number_of_elements: int = 1,
        start_index: int = 1,
    ) -> None:
        """Initialize CEMIMProp object."""
        self.object_type = object_type
        self.object_instance = object_instance
        self.property_id = (
            property_id.value
            if isinstance(property_id, ResourcePropertyId)
            else property_id
        )
        self.number_of_elements = number_of_elements
        self.start_index = start_index

    def calculated_length(self) -> int:
        """Get length of CEMI data."""
        return CEMIMPropInfo.LENGTH

    @staticmethod
    def from_knx(raw: bytes) -> CEMIMPropInfo:
        """Parse/deserialize from KNX/IP raw data."""
        if len(raw) != CEMIMPropInfo.LENGTH:
            raise CouldNotParseCEMI(
                f"Invalid CEMI length: {len(raw)}; CEMI: {raw.hex()}"
            )

        try:
            object_type = ResourceObjectType(int.from_bytes(raw[0:2], "big"))
        except ValueError:
            raise UnsupportedCEMIMessage(
                f"CEMIMProp Object Type not supported: {raw[0:2].hex()} in CEMI: {raw.hex()}"
            ) from None

        return CEMIMPropInfo(
            object_type=object_type,
            object_instance=raw[2],
            property_id=raw[3],
            number_of_elements=(raw[4] >> 4),
            start_index=(int.from_bytes(raw[4:6], "big") % 0x1000),
        )

    def to_knx(self) -> bytes:
        """Serialize to CEMI raw data."""
        return (
            self.object_type.value.to_bytes(2, "big")
            + self.object_instance.to_bytes(1, "big")
            + self.property_id.to_bytes(1, "big")
            + ((self.number_of_elements << 12) + self.start_index).to_bytes(2, "big")
        )

    def __repr__(self) -> str:
        """Return object as readable string."""
        return (
            f'object_type="{self.object_type.value}" '
            f'object_instance="{self.object_instance}" '
            f'property_id="{self.property_id}" '
            f'number_of_elements="{self.number_of_elements}" '
            f'start_index="{self.start_index}" '
        )

    def __eq__(self, other: object) -> bool:
        """Equal operator."""
        return self.__dict__ == other.__dict__


class CEMIMPropReadRequest(CEMIData):
    """Representation of CEMI Device Management Property Read Request."""

    def __init__(self, *, property_info: CEMIMPropInfo) -> None:
        """Initialize CEMIMPropReadRequest object."""
        self.property_info = property_info

    def calculated_length(self) -> int:
        """Get length of CEMI data."""
        return self.property_info.calculated_length()

    def to_knx(self) -> bytes:
        """Serialize to CEMI raw data."""
        return self.property_info.to_knx()

    @classmethod
    def from_knx(cls, raw: bytes) -> CEMIData:
        """Parse/deserialize from KNX/IP raw data."""
        return cls(property_info=CEMIMPropInfo.from_knx(raw))

    def __repr__(self) -> str:
        """Return object as readable string."""
        return f"CEMIMPropReadRequest({self.property_info})"


class CEMIMPropReadResponse(CEMIData):
    """Representation of CEMI Device Management Property Read Response."""

    def __init__(
        self,
        *,
        property_info: CEMIMPropInfo,
        data: bytes,
    ) -> None:
        """Initialize CEMIMPropReadResponse object."""
        self.property_info = property_info
        self.data = data

    @property
    def error_code(self) -> CEMIErrorCode | None:
        """Return an optional CEMI error code."""
        if self.property_info.number_of_elements == 0:
            return CEMIErrorCode(self.data[0])

        return None

    def calculated_length(self) -> int:
        """Get length of CEMI data."""
        return CEMIMPropInfo.LENGTH + len(self.data)

    def to_knx(self) -> bytes:
        """Serialize to CEMI raw data."""
        return self.property_info.to_knx() + self.data

    @classmethod
    def from_knx(cls, raw: bytes) -> CEMIData:
        """Parse/deserialize from KNX/IP raw data."""
        if len(raw) <= CEMIMPropInfo.LENGTH:
            raise CouldNotParseCEMI(
                f"CEMI Property Read Response too small. Length: {len(raw)}; CEMI: {raw.hex()}"
            )

        property_info = CEMIMPropInfo.from_knx(raw[0 : CEMIMPropInfo.LENGTH])

        # Did we get an error?
        if property_info.number_of_elements == 0 and len(raw) != 7:
            raise CouldNotParseCEMI(
                f"Invalid CEMI error response length: {len(raw)}; CEMI: {raw.hex()}"
            )

        return cls(property_info=property_info, data=raw[CEMIMPropInfo.LENGTH :])

    def __repr__(self) -> str:
        """Return object as readable string."""
        _data = (
            f'data="{self.data.hex()}" '
            if self.property_info.number_of_elements != 0
            else ""
        )
        return (
            f"CEMIMPropReadResponse("
            f"{self.property_info}"
            f'error_code="{self.error_code}" '
            f"{_data})"
        )


class CEMIMPropWriteRequest(CEMIData):
    """Representation of CEMI Device Management Property Write Request."""

    def __init__(
        self,
        *,
        property_info: CEMIMPropInfo,
        data: bytes,
    ) -> None:
        """Initialize CEMIMPropWriteRequest object."""
        self.property_info = property_info
        self.data = data

    def calculated_length(self) -> int:
        """Get length of CEMI data."""
        return CEMIMPropInfo.LENGTH + len(self.data)

    def to_knx(self) -> bytes:
        """Serialize to CEMI raw data."""
        return self.property_info.to_knx() + self.data

    @classmethod
    def from_knx(cls, raw: bytes) -> CEMIData:
        """Parse/deserialize from KNX/IP raw data."""
        if len(raw) <= CEMIMPropInfo.LENGTH:
            raise CouldNotParseCEMI(
                f"CEMI Property Write Request too small. Length: {len(raw)}; CEMI: {raw.hex()}"
            )

        property_info = CEMIMPropInfo.from_knx(raw[0 : CEMIMPropInfo.LENGTH])

        return cls(property_info=property_info, data=raw[CEMIMPropInfo.LENGTH :])

    def __repr__(self) -> str:
        """Return object as readable string."""
        return f'CEMIMPropWriteRequest({self.property_info}data="{self.data.hex()}" )'


class CEMIMPropWriteResponse(CEMIData):
    """Representation of CEMI Device Management Property Write Response."""

    def __init__(
        self,
        *,
        property_info: CEMIMPropInfo,
        error_code: CEMIErrorCode | None = None,
    ) -> None:
        """Initialize CEMIMPropWriteResponse object."""
        self.property_info = property_info
        self._error_code = error_code

    def calculated_length(self) -> int:
        """Get length of CEMI data."""
        return CEMIMPropInfo.LENGTH + (1 if self._error_code else 0)

    def to_knx(self) -> bytes:
        """Serialize to CEMI raw data."""
        if self._error_code:
            return self.property_info.to_knx() + self._error_code.value.to_bytes(
                1, "big"
            )

        return self.property_info.to_knx()

    @property
    def error_code(self) -> CEMIErrorCode | None:
        """Return an optional CEMI error code."""
        return self._error_code

    @classmethod
    def from_knx(cls, raw: bytes) -> CEMIData:
        """Parse/deserialize from KNX/IP raw data."""
        if len(raw) < CEMIMPropInfo.LENGTH:
            raise CouldNotParseCEMI(
                f"CEMI Property Write Response too small. Length: {len(raw)}; CEMI: {raw.hex()}"
            )

        property_info = CEMIMPropInfo.from_knx(raw[0 : CEMIMPropInfo.LENGTH])

        # Did we get an error?
        if property_info.number_of_elements == 0:
            if len(raw) != 7:
                raise CouldNotParseCEMI(
                    f"Invalid CEMI error response length: {len(raw)}; CEMI: {raw.hex()}"
                )
            return cls(
                property_info=property_info,
                error_code=CEMIErrorCode(raw[CEMIMPropInfo.LENGTH]),
            )

        if len(raw) != CEMIMPropInfo.LENGTH:
            raise CouldNotParseCEMI(
                f"Invalid CEMI response length: {len(raw)}; CEMI: {raw.hex()}"
            )

        return cls(property_info=property_info)

    def __repr__(self) -> str:
        """Return object as readable string."""
        return (
            f"CEMIMPropWriteResponse("
            f"{self.property_info}"
            f'error_code="{self.error_code}" )'
        )


class CEMIFrame:
    """Representation of a CEMI Frame."""

    def __init__(
        self,
        *,
        code: CEMIMessageCode,
        info: CEMIInfo | None = None,
        data: CEMIData | None,
    ) -> None:
        """Initialize CEMIFrame object."""
        self.code = code
        self.info = info or CEMIInfo()
        self.data = data

    @staticmethod
    def _has_info(code: CEMIMessageCode) -> bool:
        return code not in (
            CEMIMessageCode.M_PROP_READ_REQ,
            CEMIMessageCode.M_PROP_READ_CON,
            CEMIMessageCode.M_PROP_WRITE_REQ,
            CEMIMessageCode.M_PROP_WRITE_CON,
            CEMIMessageCode.M_PROP_INFO_IND,
            CEMIMessageCode.M_FUNC_PROP_COMMAND_REQ,
            CEMIMessageCode.M_FUNC_PROP_STATE_READ_REQ,
            CEMIMessageCode.M_FUNC_PROP_COMMAND_CON,
            CEMIMessageCode.M_FUNC_PROP_STATE_READ_CON,
            CEMIMessageCode.M_RESET_REQ,
            CEMIMessageCode.M_RESET_IND,
        )

    @staticmethod
    def _has_data(code: CEMIMessageCode) -> bool:
        return code not in (
            CEMIMessageCode.M_RESET_REQ,
            CEMIMessageCode.M_RESET_IND,
        )

    def calculated_length(self) -> int:
        """Get length of KNX/IP body."""
        length = 1

        if self._has_info(self.code):
            length += self.info.calculated_length()

        if self._has_data(self.code):
            if self.data is None:
                raise UnsupportedCEMIMessage(
                    f"CEMI data required for code: {self.code}"
                )
            length += self.data.calculated_length()

        return length

    @staticmethod
    def from_knx(raw: bytes) -> CEMIFrame:
        """Parse/deserialize from KNX/IP raw data."""
        try:
            code = CEMIMessageCode(raw[0])
        except ValueError:
            raise UnsupportedCEMIMessage(
                f"CEMIMessageCode not implemented: {raw[0]} in CEMI: {raw.hex()}"
            ) from None

        if code in (
            CEMIMessageCode.L_DATA_IND,
            CEMIMessageCode.L_DATA_REQ,
            CEMIMessageCode.L_DATA_CON,
        ):
            info, remainder = CEMIInfo.from_knx(raw[1:])
            return CEMIFrame(
                code=code,
                info=info,
                data=CEMILData.from_knx(remainder),
            )
        if code == CEMIMessageCode.M_PROP_READ_REQ:
            return CEMIFrame(code=code, data=CEMIMPropReadRequest.from_knx(raw[1:]))
        if code == CEMIMessageCode.M_PROP_READ_CON:
            return CEMIFrame(code=code, data=CEMIMPropReadResponse.from_knx(raw[1:]))
        if code == CEMIMessageCode.M_PROP_WRITE_REQ:
            return CEMIFrame(code=code, data=CEMIMPropWriteRequest.from_knx(raw[1:]))
        if code == CEMIMessageCode.M_PROP_WRITE_CON:
            return CEMIFrame(code=code, data=CEMIMPropWriteResponse.from_knx(raw[1:]))

        raise UnsupportedCEMIMessage(
            f"Could not handle CEMIMessageCode: {code} / {raw[0]} in CEMI: {raw.hex()}"
        )

    def to_knx(self) -> bytes:
        """Serialize to KNX/IP raw data."""
        raw = bytes((self.code.value,))

        if self._has_info(self.code):
            raw += self.info.to_knx()

        if self._has_data(self.code):
            if self.data is None:
                raise UnsupportedCEMIMessage(
                    f"CEMI data required for code: {self.code}"
                )
            raw += self.data.to_knx()

        return raw

    def __repr__(self) -> str:
        """Return object as readable string."""
        _info = f'info="{self.info.__repr__()}" ' if self._has_info(self.code) else ""
        _data = f'data="{self.data.__repr__()}" ' if self._has_data(self.code) else ""

        return f'<CEMIFrame code="{self.code.name}" {_info}{_data}/>'

    def __eq__(self, other: object) -> bool:
        """Equal operator."""
        return self.__dict__ == other.__dict__



"""Unit test for RemoteValueByLength objects."""

import pytest

from xknx import XKNX
from xknx.dpt import (
    DPTArray,
    DPTBase,
    DPTBinary,
    DPTColorXYY,
    DPTOpenClose,
    DPTPressure,
    DPTPressure2Byte,
    DPTTemperature,
    DPTValue1Count,
)
from xknx.exceptions import ConversionError, CouldNotParseTelegram
from xknx.remote_value.remote_value_by_length import RemoteValueByLength


class TestRemoteValueByLength:
    """Test class for RemoteValueByLength objects."""

    @pytest.mark.parametrize(
        "dpt_classes",
        [
            (DPTOpenClose, DPTPressure),  # DPTBinary payload invalid
            (DPTTemperature, DPTPressure2Byte),  # similar payload_length
            (DPTColorXYY, DPTValue1Count),  # non-numeric DPT
        ],
    )
    def test_invalid_dpt_classes(self, dpt_classes: tuple[type[DPTBase]]) -> None:
        """Test if invalid DPT classes raise ConversionError."""
        xknx = XKNX()
        with pytest.raises(ConversionError):
            RemoteValueByLength(xknx, dpt_classes=dpt_classes)  # type: ignore[arg-type]

    @pytest.mark.parametrize("payload", [DPTBinary(0), DPTArray((0, 1, 2))])
    def test_invalid_payload(self, payload: DPTArray | DPTBinary) -> None:
        """Test if invalid payloads raise CouldNotParseTelegram."""
        xknx = XKNX()
        remote_value = RemoteValueByLength(
            xknx=xknx,
            dpt_classes=(DPTPressure, DPTPressure2Byte),
        )
        with pytest.raises(CouldNotParseTelegram):
            remote_value.from_knx(payload)

    @pytest.mark.parametrize(
        ("first_dpt", "invalid_dpt"),
        [
            (DPTPressure, DPTPressure2Byte),
            (DPTPressure2Byte, DPTPressure),
        ],
    )
    def test_payload_valid_mode_assignment(
        self, first_dpt: type[DPTBase], invalid_dpt: type[DPTBase]
    ) -> None:
        """Test if DPT is assigned properly by payload length."""
        test_value = 1
        xknx = XKNX()
        remote_value = RemoteValueByLength(
            xknx=xknx,
            dpt_classes=(DPTPressure, DPTPressure2Byte),
        )
        first_payload = first_dpt.to_knx(test_value)
        invalid_payload = invalid_dpt.to_knx(test_value)

        assert remote_value._internal_dpt_class is None
        assert remote_value.from_knx(first_payload) == test_value
        assert remote_value._internal_dpt_class == first_dpt
        with pytest.raises(CouldNotParseTelegram):
            # other DPT is invalid now
            remote_value.from_knx(invalid_payload)
        # to_knx works when initialized
        assert remote_value.to_knx(test_value) == first_payload

    def test_to_knx_uninitialized(self) -> None:
        """Test to_knx raising ConversionError when DPT is not known."""
        xknx = XKNX()
        remote_value = RemoteValueByLength(
            xknx=xknx,
            dpt_classes=(DPTPressure, DPTPressure2Byte),
        )

        assert remote_value._internal_dpt_class is None
        with pytest.raises(ConversionError):
            remote_value.to_knx(1)



"""Implementation of Basic KNX 4-Byte signed (2's complement) values."""

from __future__ import annotations

from .dpt import DPTNumeric, DPTStructIntMixin


class DPT4ByteSigned(DPTStructIntMixin, DPTNumeric):
    """
    Abstraction for KNX 4 Byte "32-bit signed".

    DPT 13.***
    """

    dpt_main_number = 13
    dpt_sub_number: int | None = None
    value_type = "4byte_signed"
    payload_length = 4

    value_min = -2147483648
    value_max = 2147483647
    resolution = 1

    _struct_format = ">i"


class DPTValue4Count(DPT4ByteSigned):
    """DPT 13.001 DPT_Value_4_Count (pulse)."""

    dpt_main_number = 13
    dpt_sub_number = 1
    value_type = "pulse_4byte"
    unit = "counter pulses"


class DPTFlowRateM3H(DPT4ByteSigned):
    """DPT 13.002 DPT_FlowRate_m3/h (m³/h)."""

    dpt_main_number = 13
    dpt_sub_number = 2
    value_type = "flow_rate_m3h"
    unit = "m³/h"


class DPTActiveEnergy(DPT4ByteSigned):
    """DPT 13.010 DPT_ActiveEnergy (Wh)."""

    dpt_main_number = 13
    dpt_sub_number = 10
    value_type = "active_energy"
    unit = "Wh"
    ha_device_class = "energy"


class DPTApparantEnergy(DPT4ByteSigned):
    """DPT 13.011 DPT_ActiveEnergy (VAh)."""

    dpt_main_number = 13
    dpt_sub_number = 11
    value_type = "apparant_energy"
    unit = "VAh"


class DPTReactiveEnergy(DPT4ByteSigned):
    """DPT 13.012 DPT_ActiveEnergy (VARh)."""

    dpt_main_number = 13
    dpt_sub_number = 12
    value_type = "reactive_energy"
    unit = "VARh"


class DPTActiveEnergykWh(DPT4ByteSigned):
    """DPT 13.013 DPT_ActiveEnergy_kWh (kWh)."""

    dpt_main_number = 13
    dpt_sub_number = 13
    value_type = "active_energy_kwh"
    unit = "kWh"
    ha_device_class = "energy"


class DPTApparantEnergykVAh(DPT4ByteSigned):
    """DPT 13.014 DPT_ActiveEnergy_kVAh (kVAh)."""

    dpt_main_number = 13
    dpt_sub_number = 14
    value_type = "apparant_energy_kvah"
    unit = "kVAh"


class DPTReactiveEnergykVARh(DPT4ByteSigned):
    """DPT 13.015 DPT_ActiveEnergy (kVARh)."""

    dpt_main_number = 13
    dpt_sub_number = 15
    value_type = "reactive_energy_kvarh"
    unit = "kVARh"


class DPTActiveEnergyMWh(DPT4ByteSigned):
    """DPT 13.016 DPT_ActiveEnergy_MWh (MWh)."""

    dpt_main_number = 13
    dpt_sub_number = 16
    value_type = "active_energy_mwh"
    unit = "MWh"


class DPTLongDeltaTimeSec(DPT4ByteSigned):
    """DPT 13.100 DPT_LongDeltaTimeSec (s)."""

    dpt_main_number = 13
    dpt_sub_number = 100
    value_type = "long_delta_timesec"
    unit = "s"


class DPTDeltaVolumeLiquidLitre(DPT4ByteSigned):
    """DPT 13.1200 DPT_DeltaVolumeLiquid_Litre (water/gas total consumption)."""

    dpt_main_number = 13
    dpt_sub_number = 1200
    value_type = "delta_volume_liquid_litre"
    unit = "L"


class DPTDeltaVolumeM3(DPT4ByteSigned):
    """DPT 13.1201 DPT_DeltaVolume_m3 (water/gas total consumption volume)."""

    dpt_main_number = 13
    dpt_sub_number = 1201
    value_type = "delta_volume_m3"
    unit = "m³"



"""Unit test for KNX DPT HVAC Operation modes."""

from typing import Any

import pytest

from xknx.dpt import DPTArray, DPTHVACContrMode, DPTHVACMode, DPTHVACStatus
from xknx.dpt.dpt_20 import HeatCool, HVACControllerMode, HVACOperationMode, HVACStatus
from xknx.exceptions import ConversionError, CouldNotParseTelegram


class TestDPTHVACMode:
    """Test class for KNX DPT HVAC Operation modes."""

    def test_mode_to_knx(self) -> None:
        """Test parsing DPTHVACMode to KNX."""
        assert DPTHVACMode.to_knx(HVACOperationMode.AUTO) == DPTArray((0x00,))
        assert DPTHVACMode.to_knx(HVACOperationMode.COMFORT) == DPTArray((0x01,))
        assert DPTHVACMode.to_knx(HVACOperationMode.STANDBY) == DPTArray((0x02,))
        assert DPTHVACMode.to_knx(HVACOperationMode.ECONOMY) == DPTArray((0x03,))
        assert DPTHVACMode.to_knx(HVACOperationMode.BUILDING_PROTECTION) == DPTArray(
            (0x04,)
        )

    def test_mode_to_knx_by_string(self) -> None:
        """Test parsing DPTHVACMode string values to KNX."""
        assert DPTHVACMode.to_knx("auto") == DPTArray((0x00,))
        assert DPTHVACMode.to_knx("Comfort") == DPTArray((0x01,))
        assert DPTHVACMode.to_knx("standby") == DPTArray((0x02,))
        assert DPTHVACMode.to_knx("ECONOMY") == DPTArray((0x03,))
        assert DPTHVACMode.to_knx("Building_Protection") == DPTArray((0x04,))

    def test_mode_to_knx_wrong_value(self) -> None:
        """Test serializing DPTHVACMode to KNX with wrong value."""
        with pytest.raises(ConversionError):
            DPTHVACMode.to_knx(5)

    def test_mode_from_knx(self) -> None:
        """Test parsing DPTHVACMode from KNX."""
        assert DPTHVACMode.from_knx(DPTArray((0x00,))) == HVACOperationMode.AUTO
        assert DPTHVACMode.from_knx(DPTArray((0x01,))) == HVACOperationMode.COMFORT
        assert DPTHVACMode.from_knx(DPTArray((0x02,))) == HVACOperationMode.STANDBY
        assert DPTHVACMode.from_knx(DPTArray((0x03,))) == HVACOperationMode.ECONOMY
        assert (
            DPTHVACMode.from_knx(DPTArray((0x04,)))
            == HVACOperationMode.BUILDING_PROTECTION
        )

    def test_mode_from_knx_wrong_value(self) -> None:
        """Test parsing of DPTHVACMode with wrong value)."""
        with pytest.raises(CouldNotParseTelegram):
            DPTHVACMode.from_knx(DPTArray((1, 2)))
        with pytest.raises(ConversionError):
            DPTHVACMode.from_knx(DPTArray((0x05,)))


class TestDPTHVACControllerMode:
    """Test class for KNX DPT HVAC Controller modes."""

    def test_mode_to_knx(self) -> None:
        """Test parsing DPTHVACContrMode to KNX."""
        assert DPTHVACContrMode.to_knx(HVACControllerMode.AUTO) == DPTArray((0x00,))
        assert DPTHVACContrMode.to_knx(HVACControllerMode.HEAT) == DPTArray((0x01,))
        assert DPTHVACContrMode.to_knx(HVACControllerMode.COOL) == DPTArray((0x03,))
        assert DPTHVACContrMode.to_knx(HVACControllerMode.OFF) == DPTArray((0x06,))
        assert DPTHVACContrMode.to_knx(HVACControllerMode.DEHUMIDIFICATION) == DPTArray(
            (0x0E,)
        )

    def test_mode_to_knx_by_string(self) -> None:
        """Test parsing DPTHVACMode string values to KNX."""
        assert DPTHVACContrMode.to_knx("morning_warmup") == DPTArray((0x02,))
        assert DPTHVACContrMode.to_knx("NIGHT_PURGE") == DPTArray((0x04,))
        assert DPTHVACContrMode.to_knx("precool") == DPTArray((0x05,))
        assert DPTHVACContrMode.to_knx("Test") == DPTArray((0x07,))
        assert DPTHVACContrMode.to_knx("NODEM") == DPTArray((0x14,))

    def test_mode_to_knx_wrong_value(self) -> None:
        """Test serializing DPTHVACMode to KNX with wrong value."""
        with pytest.raises(ConversionError):
            DPTHVACContrMode.to_knx(18)
        with pytest.raises(ConversionError):
            DPTHVACContrMode.to_knx(19)
        with pytest.raises(ConversionError):
            DPTHVACContrMode.to_knx(21)

    @pytest.mark.parametrize(
        ("raw", "value"),
        [
            (0x00, HVACControllerMode.AUTO),
            (0x08, HVACControllerMode.EMERGENCY_HEAT),
            (0x09, HVACControllerMode.FAN_ONLY),
            (0x0A, HVACControllerMode.FREE_COOL),
            (0x0B, HVACControllerMode.ICE),
        ],
    )
    def test_mode_from_knx(self, raw: int, value: HVACControllerMode) -> None:
        """Test parsing DPTHVACMode from KNX."""
        assert DPTHVACContrMode.from_knx(DPTArray((raw,))) is value

    def test_mode_from_knx_wrong_value(self) -> None:
        """Test parsing of DPTHVACMode with wrong value)."""
        with pytest.raises(CouldNotParseTelegram):
            DPTHVACContrMode.from_knx(DPTArray((1, 2)))
        with pytest.raises(ConversionError):
            DPTHVACContrMode.from_knx(DPTArray((18,)))


class TestHVACStatus:
    """Test HVACStatus class."""

    @pytest.mark.parametrize(
        ("data", "dict_value"),
        [
            (
                HVACStatus(
                    mode=HVACOperationMode.COMFORT,
                    dew_point=False,
                    heat_cool=HeatCool.HEAT,
                    inactive=False,
                    frost_alarm=False,
                ),
                {
                    "mode": "comfort",
                    "dew_point": False,
                    "heat_cool": "heat",
                    "inactive": False,
                    "frost_alarm": False,
                },
            ),
            (
                HVACStatus(
                    mode=HVACOperationMode.STANDBY,
                    dew_point=False,
                    heat_cool=HeatCool.COOL,
                    inactive=True,
                    frost_alarm=False,
                ),
                {
                    "mode": "standby",
                    "dew_point": False,
                    "heat_cool": "cool",
                    "inactive": True,
                    "frost_alarm": False,
                },
            ),
        ],
    )
    def test_dict(self, data: HVACStatus, dict_value: dict[str, Any]) -> None:
        """Test from_dict and as_dict methods."""
        assert HVACStatus.from_dict(dict_value) == data
        assert data.as_dict() == dict_value

    @pytest.mark.parametrize(
        "data",
        [
            {
                "mode": 1,  # invalid
                "dew_point": False,
                "heat_cool": "heat",
                "inactive": False,
                "frost_alarm": False,
            },
            {
                "mode": "comfort",
                "dew_point": False,
                "heat_cool": "invalid",  # invalid
                "inactive": False,
                "frost_alarm": False,
            },
            {
                "mode": "comfort",
                "dew_point": False,
                "heat_cool": "nodem",  # invalid for HVACStatus
                "inactive": False,
                "frost_alarm": False,
            },
            {
                "mode": "comfort",
                "dew_point": 20,  # invalid
                "heat_cool": "heat",
                "inactive": False,
                "frost_alarm": False,
            },
        ],
    )
    def test_dict_invalid(self, data: dict[str, Any]) -> None:
        """Test from_dict with invalid data."""
        with pytest.raises(ValueError):
            HVACStatus.from_dict(data)


class TestDPTHVACStatus:
    """Test class for KNX DPTHVACStatus objects."""

    @pytest.mark.parametrize(
        ("value", "raw"),
        [
            (
                HVACStatus(
                    mode=HVACOperationMode.COMFORT,
                    dew_point=False,
                    heat_cool=HeatCool.HEAT,
                    inactive=False,
                    frost_alarm=False,
                ),
                (0x21,),
            ),
            (
                HVACStatus(
                    mode=HVACOperationMode.COMFORT,
                    dew_point=False,
                    heat_cool=HeatCool.COOL,
                    inactive=False,
                    frost_alarm=False,
                ),
                (0x01,),
            ),
            (
                HVACStatus(
                    mode=HVACOperationMode.ECONOMY,
                    dew_point=True,
                    heat_cool=HeatCool.COOL,
                    inactive=False,
                    frost_alarm=True,
                ),
                (0x94,),
            ),
        ],
    )
    def test_dpt_encoding_decoding(self, value: HVACStatus, raw: tuple[int]) -> None:
        """Test DPTHVACStatus parsing and streaming."""
        knx_value = DPTHVACStatus.to_knx(value)
        assert knx_value == DPTArray(raw)
        assert DPTHVACStatus.from_knx(knx_value) == value

    @pytest.mark.parametrize(
        ("value", "raw"),
        [
            (
                {
                    "mode": "comfort",
                    "dew_point": False,
                    "heat_cool": "heat",
                    "inactive": False,
                    "frost_alarm": False,
                },
                (0x21,),
            ),
            (
                {
                    "mode": "standby",
                    "dew_point": False,
                    "heat_cool": "cool",
                    "inactive": True,
                    "frost_alarm": False,
                },
                (0x42,),
            ),
        ],
    )
    def test_dpt_to_knx_from_dict(self, value: dict[str, Any], raw: tuple[int]) -> None:
        """Test DPTHVACStatus parsing from a dict."""
        knx_value = DPTHVACStatus.to_knx(value)
        assert knx_value == DPTArray(raw)

    @pytest.mark.parametrize(
        "value",
        [
            {"mode": "comfort", "dew_point": False, "heat_cool": "heat"},
            1,
            "cool",
        ],
    )
    def test_dpt_wrong_value_to_knx(self, value: Any) -> None:
        """Test DPTHVACStatus parsing with wrong value."""
        with pytest.raises(ConversionError):
            DPTHVACStatus.to_knx(value)

    def test_dpt_wrong_value_from_knx(self) -> None:
        """Test DPTHVACStatus parsing with wrong value."""
        with pytest.raises(CouldNotParseTelegram):
            DPTHVACStatus.from_knx(DPTArray((0xFF, 0x4E)))



"""Unit test for RemoteValueSensor objects."""

import pytest

from xknx import XKNX
from xknx.dpt import DPTArray, DPTBase, DPTBinary, DPTValue1Ucount
from xknx.exceptions import ConversionError, CouldNotParseTelegram
from xknx.remote_value import RemoteValueNumeric, RemoteValueSensor


class TestRemoteValueSensor:
    """Test class for RemoteValueSensor objects."""

    def test_value_type(self) -> None:
        """Test initializing a value_type."""
        xknx = XKNX()
        assert RemoteValueSensor(xknx=xknx, value_type="pulse")
        assert RemoteValueSensor(xknx=xknx, value_type=9)
        assert RemoteValueSensor(xknx=xknx, value_type="9.021")
        assert RemoteValueSensor(xknx=xknx, value_type="string")

    def test_wrong_value_type(self) -> None:
        """Test initializing with wrong value_type."""
        xknx = XKNX()
        with pytest.raises(ConversionError):
            RemoteValueSensor(xknx=xknx, value_type="wrong_value_type")
        with pytest.raises(ConversionError):
            RemoteValueSensor(xknx=xknx, value_type="binary")
        with pytest.raises(ConversionError):
            RemoteValueSensor(xknx=xknx, value_type=1)
        with pytest.raises(ConversionError):
            RemoteValueSensor(xknx=xknx, value_type=2)
        with pytest.raises(ConversionError):
            RemoteValueSensor(xknx=xknx, value_type=None)
        with pytest.raises(ConversionError):
            RemoteValueSensor(xknx=xknx)

    def test_payload_length_defined(self) -> None:
        """Test if all members of DPTMAP implement payload_length."""
        for dpt_class in DPTBase.__recursive_subclasses__():
            assert isinstance(dpt_class.payload_length, int)

    def test_payload_invalid(self) -> None:
        """Test invalid payloads."""
        xknx = XKNX()
        remote_value = RemoteValueSensor(xknx=xknx, value_type="pulse")

        assert remote_value.dpt_class == DPTValue1Ucount
        with pytest.raises(CouldNotParseTelegram):
            remote_value.from_knx(None)
        with pytest.raises(CouldNotParseTelegram):
            remote_value.from_knx(DPTArray((1, 2, 3, 4)))
        with pytest.raises(CouldNotParseTelegram):
            remote_value.from_knx(DPTBinary(1))


class TestRemoteValueNumeric:
    """Test class for RemoteValueNumeric objects."""

    def test_value_type(self) -> None:
        """Test initializing a value_type."""
        xknx = XKNX()
        assert RemoteValueNumeric(xknx=xknx, value_type="pulse")
        assert RemoteValueNumeric(xknx=xknx, value_type=9)
        assert RemoteValueNumeric(xknx=xknx, value_type="9.021")

    def test_wrong_value_type(self) -> None:
        """Test initializing with wrong value_type."""
        xknx = XKNX()
        with pytest.raises(ConversionError):
            RemoteValueNumeric(xknx=xknx, value_type="string")
        with pytest.raises(ConversionError):
            RemoteValueNumeric(xknx=xknx, value_type=16)
        with pytest.raises(ConversionError):
            RemoteValueNumeric(xknx=xknx, value_type="binary")
        with pytest.raises(ConversionError):
            RemoteValueNumeric(xknx=xknx)



"""Module for KNX Data Secure."""

from __future__ import annotations

from collections.abc import Iterator
from contextlib import contextmanager
from copy import copy
from datetime import datetime, timezone
import logging
import time

from xknx.cemi import CEMILData
from xknx.exceptions import DataSecureError
from xknx.telegram.address import GroupAddress, IndividualAddress
from xknx.telegram.apci import APCI, SecureAPDU

from .data_secure_asdu import (
    SecureData,
    SecurityAlgorithmIdentifier,
    SecurityALService,
    SecurityControlField,
)
from .keyring import Keyring

_LOGGER = logging.getLogger("xknx.data_secure")

# Same timedelta in milliseconds as used in Falcon used for initial sequence_number_sending
# py3.10 backwards compatibility - py3.11 "2018-01-05T00:00:00Z" is supported
_SEQUENCE_NUMBER_INIT_TIMESTAMP = datetime.fromisoformat(
    "2018-01-05T00:00:00+00:00"
).timestamp()


def _initial_sequence_number() -> int:
    """Return an initial sequence number for sending Data Secure Telegrams."""
    return int((time.time() - _SEQUENCE_NUMBER_INIT_TIMESTAMP) * 1000)


def is_data_secure(cemi_data: CEMILData) -> bool:
    """Return True if the payload frame is DataSecure encrypted / authenticated."""
    return isinstance(cemi_data.payload, SecureAPDU)


class DataSecure:
    """Class for KNX Data Secure handling."""

    __slots__ = (
        "_group_key_table",
        "_individual_address_table",
        "_sequence_number_sending",
    )

    def __init__(
        self,
        *,
        group_key_table: dict[GroupAddress, bytes],
        individual_address_table: dict[IndividualAddress, int],
        last_sequence_number_sending: int | None = None,
    ) -> None:
        """Initialize DataSecure class."""
        self._group_key_table = group_key_table
        self._individual_address_table = individual_address_table
        self._sequence_number_sending = (
            last_sequence_number_sending or _initial_sequence_number()
        )
        # Holds the last valid sequence number for each individual address.
        # Use sequence_number from keyfile as initial value or 0 from senders for all IAs ?

        if not 0 < self._sequence_number_sending < 0xFFFFFFFFFFFF:
            _local_time_info = (
                f" Local time not set properly? {datetime.now(timezone.utc).isoformat()}"
                if not last_sequence_number_sending
                else ""
            )
            raise DataSecureError(
                f"Initial sequence number out of range: {self._sequence_number_sending}"
                f"{_local_time_info}"
            )
        _LOGGER.info(
            "Data Secure initialized for %s group addresses from %s individual addresses.",
            len(self._group_key_table),
            len(self._individual_address_table),
        )
        _LOGGER.debug(
            "Data Secure initial sequence number: %s, groups: %s, senders: %s",
            self._sequence_number_sending,
            [str(ga) for ga in self._group_key_table],
            [str(ia) for ia in self._individual_address_table],
        )

    @staticmethod
    def init_from_keyring(keyring: Keyring) -> DataSecure | None:
        """
        Initialize DataSecure from Keyring.

        Return None if no Data Secure information is found in the Keyring.
        """
        ga_key_table = keyring.get_data_secure_group_keys()
        ia_seq_table = keyring.get_data_secure_senders()
        # TODO: persist local individual_address_table and update from that file on start
        #       to have more fresh initial sequence numbers
        if not ga_key_table:
            return None
        return DataSecure(
            group_key_table=ga_key_table,
            individual_address_table=ia_seq_table,
        )

    def get_sequence_number(self) -> int:
        """Return current sequence number sending and increment local stored value."""
        seq_nr = self._sequence_number_sending
        self._sequence_number_sending += 1
        return seq_nr

    @contextmanager
    def check_sequence_number(
        self, source_address: IndividualAddress, received_sequence_number: int
    ) -> Iterator[None]:
        """
        Check the last valid sequence number for incoming frames from `source_address`.

        Update the Security Individual Address Table if no further exception is raised.
        Raise `DataSecureError` if sequence number is invalid or sender is not known.
        """
        try:
            last_valid_sequence_number = self._individual_address_table[source_address]
        except KeyError:
            raise DataSecureError(
                f"Source address not found in Security Individual Address Table: {source_address}",
                log_level=logging.INFO,
            ) from None
        if not received_sequence_number > last_valid_sequence_number:
            # TODO: implement and increment Security Failure Log counter (not when equal)
            raise DataSecureError(
                f"Sequence number too low for {source_address}: "
                f"{received_sequence_number} received, {last_valid_sequence_number} last valid",
                log_level=logging.WARNING,
            )

        yield
        # Don't increment sequence number if exception is raised while decrypting (yield)
        self._individual_address_table[source_address] = received_sequence_number

    def received_cemi(self, cemi_data: CEMILData) -> CEMILData:
        """Handle received CEMI frame."""
        # Data Secure frame
        if isinstance(cemi_data.payload, SecureAPDU):
            return self._received_secure_cemi(cemi_data, cemi_data.payload)
        # Plain group communication frame
        if isinstance(cemi_data.dst_addr, GroupAddress):
            if cemi_data.dst_addr in self._group_key_table:
                raise DataSecureError(
                    f"Discarding frame with plain APDU for secure group address: {cemi_data}",
                    log_level=logging.WARNING,
                )
            return cemi_data
        # Plain point-to-point frame
        #   No point to point key table is implemented at the moment as ETS can't even configure this
        #   only way to communicate point-to-point with data secure currently is with tool key
        #   - which we don't have
        return cemi_data

    def _received_secure_cemi(
        self, cemi_data: CEMILData, s_apdu: SecureAPDU
    ) -> CEMILData:
        """Handle received secured CEMI frame."""
        if s_apdu.scf.service is not SecurityALService.S_A_DATA:
            raise DataSecureError(
                f"Only SecurityALService.S_A_DATA supported {cemi_data}",
                log_level=logging.DEBUG,
            )
        if s_apdu.scf.system_broadcast or s_apdu.scf.tool_access:
            # TODO: handle incoming responses with tool key of sending device
            # when we can send with tool key
            raise DataSecureError(
                f"System broadcast and tool access not supported {cemi_data}",
                log_level=logging.DEBUG,
            )

        # Secure group communication frame
        if isinstance(cemi_data.dst_addr, GroupAddress):
            if not (key := self._group_key_table.get(cemi_data.dst_addr)):
                raise DataSecureError(
                    f"No key found for group address {cemi_data.dst_addr} from {cemi_data.src_addr}",
                    log_level=logging.INFO,
                )
        # Secure point-to-point frame
        else:
            # TODO: maybe possible to implement this over tool key
            raise DataSecureError(
                f"Secure Point-to-Point communication not supported {cemi_data}",
                log_level=logging.DEBUG,
            )

        with self.check_sequence_number(
            source_address=cemi_data.src_addr,
            received_sequence_number=int.from_bytes(
                s_apdu.secured_data.sequence_number_bytes, "big"
            ),
        ):
            _address_fields_raw = (
                cemi_data.src_addr.to_knx() + cemi_data.dst_addr.to_knx()
            )
            plain_apdu_raw = s_apdu.secured_data.get_plain_apdu(
                key=key,
                scf=s_apdu.scf,
                address_fields_raw=_address_fields_raw,
                frame_flags=cemi_data.flags,
                tpci=cemi_data.tpci,
            )
        decrypted_payload = APCI.from_knx(plain_apdu_raw)
        _LOGGER.debug("Unpacked APDU %s from %s", decrypted_payload, s_apdu)

        plain_cemi_data = copy(cemi_data)
        plain_cemi_data.payload = decrypted_payload
        return plain_cemi_data

    def outgoing_cemi(self, cemi_data: CEMILData) -> CEMILData:
        """Handle outgoing CEMI frame. Pass through as plain frame or encrypt."""
        # Outgoing  group communication frame
        if isinstance(cemi_data.dst_addr, GroupAddress):
            if key := self._group_key_table.get(cemi_data.dst_addr):
                scf = SecurityControlField(
                    algorithm=SecurityAlgorithmIdentifier.CCM_ENCRYPTION,
                    service=SecurityALService.S_A_DATA,
                    system_broadcast=False,
                    tool_access=False,
                )
                return self._secure_data_cemi(key=key, scf=scf, cemi_data=cemi_data)
            return cemi_data
        # Outgoing secure point-to-point frames are sent plain.
        # Data Secure point-to-point is not supported.
        return cemi_data

    def _secure_data_cemi(
        self,
        key: bytes,
        scf: SecurityControlField,
        cemi_data: CEMILData,
    ) -> CEMILData:
        """Wrap encrypted payload of a plain CEMILData in a SecureAPDU."""
        plain_apdu_raw: bytes | bytearray

        if cemi_data.payload is not None:
            plain_apdu_raw = cemi_data.payload.to_knx()
        else:
            # TODO: test if this is correct
            plain_apdu_raw = b""  # used in point-to-point eg. TConnect
        secure_asdu = SecureData.init_from_plain_apdu(
            key=key,
            apdu=plain_apdu_raw,
            scf=scf,
            sequence_number=self.get_sequence_number(),
            address_fields_raw=cemi_data.src_addr.to_knx()
            + cemi_data.dst_addr.to_knx(),
            frame_flags=cemi_data.flags,
            tpci=cemi_data.tpci,
        )
        secure_cemi_data = copy(cemi_data)
        secure_cemi_data.payload = SecureAPDU(scf=scf, secured_data=secure_asdu)
        _LOGGER.debug(
            "Secured APDU %s with %s", cemi_data.payload, secure_cemi_data.payload
        )
        return secure_cemi_data



"""Unit test for RemoteValueColorRGB objects."""

import pytest

from xknx import XKNX
from xknx.dpt import DPTArray, DPTBinary, RGBColor
from xknx.exceptions import ConversionError
from xknx.remote_value import RemoteValueColorRGB
from xknx.telegram import GroupAddress, Telegram
from xknx.telegram.apci import GroupValueWrite


class TestRemoteValueColorRGB:
    """Test class for RemoteValueColorRGB objects."""

    def test_to_knx(self) -> None:
        """Test to_knx function with normal operation."""
        xknx = XKNX()
        remote_value = RemoteValueColorRGB(xknx)
        assert remote_value.to_knx(RGBColor(100, 101, 102)) == DPTArray(
            (0x64, 0x65, 0x66)
        )
        assert remote_value.to_knx(RGBColor(100, 101, 102)) == DPTArray(
            (0x64, 0x65, 0x66)
        )

    def test_from_knx(self) -> None:
        """Test from_knx function with normal operation."""
        xknx = XKNX()
        remote_value = RemoteValueColorRGB(xknx)
        assert remote_value.from_knx(DPTArray((0x64, 0x65, 0x66))) == RGBColor(
            100, 101, 102
        )

    def test_to_knx_error(self) -> None:
        """Test to_knx function with wrong parametern."""
        xknx = XKNX()
        remote_value = RemoteValueColorRGB(xknx)
        with pytest.raises(ConversionError):
            remote_value.to_knx(RGBColor(100, 101, 256))
        with pytest.raises(ConversionError):
            remote_value.to_knx(RGBColor(100, -1, 102))
        with pytest.raises(ConversionError):
            remote_value.to_knx(RGBColor("100", 101, 102))

    async def test_set(self) -> None:
        """Test setting value."""
        xknx = XKNX()
        remote_value = RemoteValueColorRGB(xknx, group_address=GroupAddress("1/2/3"))
        remote_value.set(RGBColor(100, 101, 102))
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTArray((0x64, 0x65, 0x66))),
        )
        remote_value.set(RGBColor(100, 101, 104))
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTArray((0x64, 0x65, 0x68))),
        )

    def test_process(self) -> None:
        """Test process telegram."""
        xknx = XKNX()
        remote_value = RemoteValueColorRGB(xknx, group_address=GroupAddress("1/2/3"))
        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTArray((0x64, 0x65, 0x66))),
        )
        remote_value.process(telegram)
        assert remote_value.value == RGBColor(100, 101, 102)

    def test_to_process_error(self) -> None:
        """Test process erroneous telegram."""
        xknx = XKNX()
        remote_value = RemoteValueColorRGB(xknx, group_address=GroupAddress("1/2/3"))

        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTBinary(1)),
        )
        assert remote_value.process(telegram) is False

        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTArray((0x64, 0x65, 0x66, 0x67))),
        )
        assert remote_value.process(telegram) is False

        assert remote_value.value is None



"""Unit test for KNX/IP Connect Request/Response."""

from unittest.mock import patch

from xknx.io.request_response import Connect
from xknx.io.transport import UDPTransport
from xknx.knxip import (
    HPAI,
    ConnectionStateRequest,
    ConnectRequest,
    ConnectResponse,
    ConnectResponseData,
    ErrorCode,
    KNXIPFrame,
)
from xknx.telegram import IndividualAddress


class TestConnect:
    """Test class for xknx/io/Connect objects."""

    async def test_connect(self) -> None:
        """Test connecting from KNX bus."""
        udp_transport = UDPTransport(("192.168.1.1", 0), ("192.168.1.2", 1234))
        local_hpai = HPAI(ip_addr="192.168.1.3", port=4321)
        connect = Connect(udp_transport, local_hpai=local_hpai)
        connect.timeout_in_seconds = 0

        assert connect.awaited_response_class == ConnectResponse

        # Expected KNX/IP-Frame:
        exp_knxipframe = KNXIPFrame.init_from_body(
            ConnectRequest(
                control_endpoint=local_hpai,
                data_endpoint=local_hpai,
            )
        )

        with (
            patch("xknx.io.transport.UDPTransport.send") as mock_udp_send,
            patch("xknx.io.transport.UDPTransport.getsockname") as mock_udp_getsockname,
        ):
            mock_udp_getsockname.return_value = ("192.168.1.3", 4321)
            await connect.start()
            mock_udp_send.assert_called_with(exp_knxipframe)

        # Response KNX/IP-Frame with wrong type
        wrong_knxipframe = KNXIPFrame.init_from_body(ConnectionStateRequest())
        with patch("logging.Logger.warning") as mock_warning:
            connect.response_rec_callback(wrong_knxipframe, HPAI(), None)
            mock_warning.assert_called_with("Could not understand knxipframe")

        # Response KNX/IP-Frame with error:
        err_knxipframe = KNXIPFrame.init_from_body(
            ConnectResponse(status_code=ErrorCode.E_CONNECTION_ID)
        )
        with patch("logging.Logger.debug") as mock_warning:
            connect.response_rec_callback(err_knxipframe, HPAI(), None)
            mock_warning.assert_called_with(
                "Error: KNX bus responded to request of type '%s' with error in '%s': %s",
                type(connect).__name__,
                type(err_knxipframe.body).__name__,
                ErrorCode.E_CONNECTION_ID,
            )

        # Correct Response KNX/IP-Frame:
        res_knxipframe = KNXIPFrame.init_from_body(
            ConnectResponse(
                communication_channel=23,
                crd=ConnectResponseData(individual_address=IndividualAddress(7)),
            )
        )
        connect.response_rec_callback(res_knxipframe, HPAI(), None)
        assert connect.success
        assert connect.communication_channel == 23
        assert connect.crd.individual_address.raw == 7

    async def test_connect_route_back_true(self) -> None:
        """Test connecting from KNX bus."""
        udp_transport = UDPTransport(("192.168.1.1", 0), ("192.168.1.2", 1234))
        local_hpai = HPAI()  # route_back: No IP address, no port, UDP
        connect = Connect(udp_transport, local_hpai=local_hpai)
        connect.timeout_in_seconds = 0

        assert connect.awaited_response_class == ConnectResponse

        # Expected KNX/IP-Frame:
        exp_knxipframe = KNXIPFrame.init_from_body(ConnectRequest())
        with (
            patch("xknx.io.transport.UDPTransport.send") as mock_udp_send,
            patch("xknx.io.transport.UDPTransport.getsockname") as mock_udp_getsockname,
        ):
            mock_udp_getsockname.return_value = ("192.168.1.3", 4321)
            await connect.start()
            mock_udp_send.assert_called_with(exp_knxipframe)

        # Response KNX/IP-Frame with wrong type
        wrong_knxipframe = KNXIPFrame.init_from_body(ConnectionStateRequest())
        with patch("logging.Logger.warning") as mock_warning:
            connect.response_rec_callback(wrong_knxipframe, HPAI(), None)
            mock_warning.assert_called_with("Could not understand knxipframe")

        # Response KNX/IP-Frame with error:
        err_knxipframe = KNXIPFrame.init_from_body(
            ConnectResponse(status_code=ErrorCode.E_CONNECTION_ID)
        )
        with patch("logging.Logger.debug") as mock_warning:
            connect.response_rec_callback(err_knxipframe, HPAI(), None)
            mock_warning.assert_called_with(
                "Error: KNX bus responded to request of type '%s' with error in '%s': %s",
                type(connect).__name__,
                type(err_knxipframe.body).__name__,
                ErrorCode.E_CONNECTION_ID,
            )

        # Correct Response KNX/IP-Frame:
        res_knxipframe = KNXIPFrame.init_from_body(
            ConnectResponse(
                communication_channel=23,
                crd=ConnectResponseData(individual_address=IndividualAddress(7)),
            )
        )
        connect.response_rec_callback(res_knxipframe, HPAI(), None)
        assert connect.success
        assert connect.communication_channel == 23
        assert connect.crd.individual_address.raw == 7



"""
Module for managing a remote date and time values.

DPT 10.001, 11.001 and 19.001
"""

from __future__ import annotations

from xknx.dpt import DPTDate, DPTDateTime, DPTTime
from xknx.dpt.dpt_10 import KNXTime
from xknx.dpt.dpt_11 import KNXDate
from xknx.dpt.dpt_19 import KNXDateTime

from .remote_value import RemoteValue


class RemoteValueTime(RemoteValue[KNXTime]):
    """Abstraction for remote value of KNX 3 octet time."""

    dpt_class = DPTTime


class RemoteValueDate(RemoteValue[KNXDate]):
    """Abstraction for remote value of KNX 3 octet date."""

    dpt_class = DPTDate


class RemoteValueDateTime(RemoteValue[KNXDateTime]):
    """Abstraction for remote value of KNX 8 octet datetime."""

    dpt_class = DPTDateTime



"""
Module for serialization and deserialization of TPCI payloads.

TPCI stands for Transport Layer Protocol Control Information.
"""

from __future__ import annotations

from abc import ABC
from typing import ClassVar

from xknx.exceptions import ConversionError

CONTROL_BIT_MASK = 0x80
NUMBERED_BIT_MASK = 0x40


class TPCI(ABC):
    """
    Base class for TCPI services.

    This base class is only the interface for the derived classes.
    """

    __slots__ = ()

    control: ClassVar[bool]
    numbered: ClassVar[bool]
    sequence_number: int = 0
    control_flags: ClassVar[int | None] = None

    def to_knx(self) -> int:
        """Serialize to KNX/IP raw data."""
        return (
            self.control << 7
            | self.numbered << 6
            | (self.sequence_number & 0xF) << 2
            | (self.control_flags or 0)
        )

    def __eq__(self, other: object) -> bool:
        """Equal operator."""
        return (
            isinstance(other, self.__class__)
            and self.sequence_number == other.sequence_number
        )

    def __repr__(self) -> str:
        """Return object as readable string."""
        _sequence_number = (
            f"sequence_number={self.sequence_number}" if self.numbered else ""
        )
        return f"{self.__class__.__name__}({_sequence_number})"

    @staticmethod
    def resolve(raw_tpci: int, dst_is_group_address: bool, dst_is_zero: bool) -> TPCI:
        """
        Return TPCI instance from TPCI command.

        See KNX Specifications 03_03_04 Transport Layer v01.02.02 AS §2 TPDU
        """
        control = raw_tpci & CONTROL_BIT_MASK
        numbered = raw_tpci & NUMBERED_BIT_MASK
        sequence_number = (raw_tpci >> 2) & 0xF

        if dst_is_group_address:
            if control or numbered:
                raise ConversionError("Invalid TPCI flags in group addressed frame.")
            if not sequence_number:
                if dst_is_zero:
                    return TDataBroadcast()
                return TDataGroup()
            if sequence_number == 1:
                # TDataTagGroup uses sequence number field as flag
                return TDataTagGroup()

        if not numbered and sequence_number:
            raise ConversionError("Sequence number not allowed for unnumbered TPCI")

        if not control:
            # data - last 2 bits are part of APCI
            if numbered:
                return TDataConnected(sequence_number=sequence_number)
            return TDataIndividual()
        # unnumbered control
        control_flags = raw_tpci & 0b11
        if not numbered:
            if control_flags == 0:
                return TConnect()
            if control_flags == 1:
                return TDisconnect()
        # numbered control
        if control_flags == 0b10:
            return TAck(sequence_number=sequence_number)
        if control_flags == 0b11:
            return TNak(sequence_number=sequence_number)

        raise ConversionError(f"Unknown TPCI {raw_tpci:#10b}.")


class TDataGroup(TPCI):
    """T_Data_Group class."""

    __slots__ = ()

    control = False
    numbered = False

    def to_knx(self) -> int:
        """Serialize to KNX/IP raw data."""
        return 0


class TDataBroadcast(TPCI):
    """T_Data_Broadcast class."""

    __slots__ = ()

    control = False
    numbered = False

    def to_knx(self) -> int:
        """Serialize to KNX/IP raw data."""
        return 0


class TDataTagGroup(TPCI):
    """T_Data_Tag_Group class."""

    __slots__ = ()

    control = False
    numbered = False
    sequence_number = 0b0001


class TDataIndividual(TPCI):
    """T_Data_Individual class."""

    __slots__ = ()

    control = False
    numbered = False

    def to_knx(self) -> int:
        """Serialize to KNX/IP raw data."""
        return 0


class TDataConnected(TPCI):
    """T_Data_Connected class."""

    __slots__ = ("sequence_number",)

    control = False
    numbered = True

    def __init__(self, sequence_number: int) -> None:
        """Initialize TDataConnected."""
        self.sequence_number = sequence_number


class TConnect(TPCI):
    """T_Connect class."""

    __slots__ = ()

    control = True
    numbered = False
    control_flags = 0b00


class TDisconnect(TPCI):
    """T_Disconnect class."""

    __slots__ = ()

    control = True
    numbered = False
    control_flags = 0b01


class TAck(TPCI):
    """T_Ack class."""

    __slots__ = ("sequence_number",)

    control = True
    numbered = True
    control_flags = 0b10

    def __init__(self, sequence_number: int) -> None:
        """Initialize TAck."""
        self.sequence_number = sequence_number


class TNak(TPCI):
    """T_Nak class."""

    __slots__ = ("sequence_number",)

    control = True
    numbered = True
    control_flags = 0b11

    def __init__(self, sequence_number: int) -> None:
        """Initialize TNak."""
        self.sequence_number = sequence_number



"""Unit test for KNX 8 byte signed objects."""

import pytest

from xknx.dpt import DPT8ByteSigned, DPTArray
from xknx.exceptions import ConversionError


class TestDPT8ByteSigned:
    """Test class for KNX 8 byte signed objects."""

    @pytest.mark.parametrize(
        ("raw", "expected"),
        (
            (b"\x00\x00\x00\x00\x00\x00\x00\x00", 0),
            (b"\x00\x00\x00\x00\x00\x00\x00\x01", 1),
            (b"\x00\x00\x00\x00\x00\x00\x00\xe6", 230),
            (b"\xff\xff\xff\xff\xff\xff\xff\x1a", -230),
            # limits
            (b"\x7f\xff\xff\xff\xff\xff\xff\xff", 9_223_372_036_854_775_807),
            (b"\x80\x00\x00\x00\x00\x00\x00\x00", -9_223_372_036_854_775_808),
        ),
    )
    def test_values(self, raw: bytes, expected: int) -> None:
        """Test valid values."""
        assert DPT8ByteSigned.to_knx(expected) == DPTArray(raw)
        assert DPT8ByteSigned.from_knx(DPTArray(raw)) == expected

    @pytest.mark.parametrize(
        "value", (9_223_372_036_854_775_808, -9_223_372_036_854_775_809)
    )
    def test_exceeding_limits(self, value: int) -> None:
        """Test invalid values."""
        with pytest.raises(ConversionError):
            DPT8ByteSigned.to_knx(value)



"""Package containing all objects for connecting to sockets."""

# ruff: noqa: F401
from .ip_transport import KNXIPTransport
from .tcp_transport import TCPTransport
from .udp_transport import UDPTransport



"""Unit test for KNX/IP Session Request/Response."""

from unittest.mock import Mock, patch

from xknx.io.request_response import Session
from xknx.knxip import HPAI, KNXIPFrame, SessionRequest, SessionResponse, SessionStatus


class TestSession:
    """Test class for xknx/io/Session objects."""

    async def test_session(self) -> None:
        """Test authenticating to secure KNX device."""
        transport_mock = Mock()
        ecdh_public_key = bytes(16)
        session = Session(
            transport_mock,
            ecdh_client_public_key=ecdh_public_key,
        )
        session.timeout_in_seconds = 0

        assert session.awaited_response_class == SessionResponse

        # Expected KNX/IP-Frame:
        exp_knxipframe = KNXIPFrame.init_from_body(
            SessionRequest(ecdh_client_public_key=ecdh_public_key)
        )

        await session.start()
        transport_mock.send.assert_called_with(exp_knxipframe)

        # Response KNX/IP-Frame with wrong type
        wrong_knxipframe = KNXIPFrame.init_from_body(SessionStatus())
        with patch("logging.Logger.warning") as mock_warning:
            session.response_rec_callback(wrong_knxipframe, HPAI(), None)
            mock_warning.assert_called_with("Could not understand knxipframe")
            assert session.success is False

        # Correct Response KNX/IP-Frame:
        res_knxipframe = KNXIPFrame.init_from_body(SessionResponse(secure_session_id=5))
        session.response_rec_callback(res_knxipframe, HPAI(), None)
        assert session.success is True
        assert session.response.secure_session_id == 5



"""
Module for Serialization and Deserialization of a KNX Tunnelling Feature service.

The Tunnelling Client shall use the Tunnelling Feature services TUNNELLING_FEATURE_GET and
TUNNELLING_FEATURE_SET to read and respectively write features related to the Tunnelling
interface and the Tunnelling host device.
The Tunnelling Server shall use the service TUNNELLING_FEATURE_INFO to spontaneously report
Tunnelling Clients about relevant changes in the state of itself or the Tunnelling connection.
"""

from __future__ import annotations

import struct

from xknx.exceptions import CouldNotParseKNXIP
from xknx.management.application_layer_enum import ReturnCode

from .body import KNXIPBody, KNXIPBodyResponse
from .error_code import ErrorCode
from .knxip_enum import KNXIPServiceType, TunnellingFeatureType


class _TunnellingFeature:
    """Base representation of a KNX Tunnelling Feature interface request."""

    HEADER_LENGTH = 4
    FEATURE_ID_LENGTH = 2

    def __init__(
        self,
        communication_channel_id: int = 1,
        sequence_counter: int = 0,
        status_code: ErrorCode = ErrorCode.E_NO_ERROR,
        feature_type: TunnellingFeatureType = TunnellingFeatureType.SUPPORTED_EMI_TYPE,
        data: bytes = b"",
    ) -> None:
        """Initialize _TunnellingFeature object."""
        self.communication_channel_id = communication_channel_id
        self.sequence_counter = sequence_counter
        self.status_code = status_code
        self.feature_type = feature_type
        self.data = data

    def _has_data(self) -> bool:
        return True

    def calculated_length(self) -> int:
        """Get length of KNX/IP body."""
        data_size = len(self.data) + (len(self.data) % 2) if self._has_data() else 0
        return (
            _TunnellingFeature.HEADER_LENGTH
            + _TunnellingFeature.FEATURE_ID_LENGTH
            + data_size
        )

    def from_knx(self, raw: bytes) -> int:
        """Parse/deserialize from KNX/IP raw data."""
        if raw[0] != _TunnellingFeature.HEADER_LENGTH:  # structure_length field
            raise CouldNotParseKNXIP("TunnellingFeature header has invalid length")
        self.communication_channel_id = raw[1]
        self.sequence_counter = raw[2]
        self.status_code = ErrorCode(raw[3])
        self.feature_type = TunnellingFeatureType(raw[4])
        self.data = raw[6:]
        if self._has_data() and len(self.data) == 0:
            raise CouldNotParseKNXIP("TunnellingFeature missing data")
        if not self._has_data() and len(self.data) > 0:
            raise CouldNotParseKNXIP("TunnellingFeature unexpected data")
        return len(raw)

    def to_knx(self) -> bytes:
        """Serialize to KNX/IP raw data."""
        data_size = len(self.data) + (len(self.data) % 2) if self._has_data() else 0
        return struct.pack(
            # Append a NUL byte if data size is uneven
            f"!BBBBBx{data_size}s",
            _TunnellingFeature.HEADER_LENGTH,
            self.communication_channel_id,
            self.sequence_counter,
            self.status_code.value,
            self.feature_type.value,
            self.data,
        )

    def __repr__(self) -> str:
        """Return object as readable string."""
        _data = f'data="{self.data.hex()}" ' if self._has_data() else ""
        return (
            f"<{self.__class__.__name__} "
            f'communication_channel_id="{self.communication_channel_id}" '
            f'sequence_counter="{self.sequence_counter}" '
            f'feature_type="{self.feature_type}" '
            f"{_data}/>"
        )


class TunnellingFeatureGet(_TunnellingFeature, KNXIPBody):
    """Representation of a KNX Tunnelling Feature Get request."""

    SERVICE_TYPE = KNXIPServiceType.TUNNELLING_FEATURE_GET

    def __init__(
        self,
        communication_channel_id: int = 1,
        sequence_counter: int = 0,
        feature_type: TunnellingFeatureType = TunnellingFeatureType.SUPPORTED_EMI_TYPE,
    ) -> None:
        """Initialize TunnellingFeatureGet object."""
        super().__init__(
            communication_channel_id=communication_channel_id,
            sequence_counter=sequence_counter,
            feature_type=feature_type,
        )

    def _has_data(self) -> bool:
        return False


class TunnellingFeatureSet(_TunnellingFeature, KNXIPBody):
    """Representation of a KNX Tunnelling Feature Set request."""

    SERVICE_TYPE = KNXIPServiceType.TUNNELLING_FEATURE_SET

    def __init__(
        self,
        communication_channel_id: int = 1,
        sequence_counter: int = 0,
        feature_type: TunnellingFeatureType = TunnellingFeatureType.SUPPORTED_EMI_TYPE,
        data: bytes = b"",
    ) -> None:
        """Initialize TunnellingFeatureSet object."""
        super().__init__(
            communication_channel_id=communication_channel_id,
            sequence_counter=sequence_counter,
            feature_type=feature_type,
            data=data,
        )


class TunnellingFeatureInfo(_TunnellingFeature, KNXIPBody):
    """Representation of a KNX Tunnelling Feature Info indication."""

    SERVICE_TYPE = KNXIPServiceType.TUNNELLING_FEATURE_INFO

    def __init__(
        self,
        communication_channel_id: int = 1,
        sequence_counter: int = 0,
        feature_type: TunnellingFeatureType = TunnellingFeatureType.SUPPORTED_EMI_TYPE,
        data: bytes = b"",
    ) -> None:
        """Initialize TunnellingFeatureInfo object."""
        super().__init__(
            communication_channel_id=communication_channel_id,
            sequence_counter=sequence_counter,
            feature_type=feature_type,
            data=data,
        )


class TunnellingFeatureResponse(_TunnellingFeature, KNXIPBodyResponse):
    """Representation of a KNX Tunnelling Feature response."""

    SERVICE_TYPE = KNXIPServiceType.TUNNELLING_FEATURE_RESPONSE

    def __init__(
        self,
        communication_channel_id: int = 1,
        sequence_counter: int = 0,
        feature_type: TunnellingFeatureType = TunnellingFeatureType.SUPPORTED_EMI_TYPE,
        return_code: ReturnCode = ReturnCode.E_SUCCESS,
        data: bytes = b"",
    ) -> None:
        """Initialize TunnellingFeatureSet object."""
        super().__init__(
            communication_channel_id=communication_channel_id,
            sequence_counter=sequence_counter,
            feature_type=feature_type,
            data=data,
        )
        self.return_code = return_code

    def from_knx(self, raw: bytes) -> int:
        """Parse/deserialize from KNX/IP raw data."""
        if raw[0] != _TunnellingFeature.HEADER_LENGTH:  # structure_length field
            raise CouldNotParseKNXIP("TunnellingFeature header has invalid length")
        self.communication_channel_id = raw[1]
        self.sequence_counter = raw[2]
        self.status_code = ErrorCode(raw[3])
        self.feature_type = TunnellingFeatureType(raw[4])
        self.return_code = ReturnCode(raw[5])
        self.data = raw[6:]
        if self.return_code is ReturnCode.E_SUCCESS and len(self.data) == 0:
            # Data may be omitted by some servers when an error occurred
            raise CouldNotParseKNXIP("TunnellingFeature missing data.")
        return len(raw)

    def to_knx(self) -> bytes:
        """Serialize to KNX/IP raw data."""
        data_size = len(self.data) + (len(self.data) % 2) if self._has_data() else 0
        return struct.pack(
            # Append a NUL byte if data size is uneven
            f"!BBBBBB{data_size}s",
            _TunnellingFeature.HEADER_LENGTH,
            self.communication_channel_id,
            self.sequence_counter,
            self.status_code.value,
            self.feature_type.value,
            self.return_code.value,
            self.data,
        )

    def __repr__(self) -> str:
        """Return object as readable string."""
        return (
            "<TunnellingFeatureResponse "
            f'communication_channel_id="{self.communication_channel_id}" '
            f'sequence_counter="{self.sequence_counter}" '
            f'status_code="{self.status_code}" '
            f'feature_type="{self.feature_type}" '
            f'return_code="{self.return_code}" '
            f'data="{self.data.hex()}" />'
        )



"""Module for Serialization and Deserialization of a KNX Device Configuration Request."""

from __future__ import annotations

from xknx.exceptions import CouldNotParseKNXIP

from .body import KNXIPBody
from .knxip_enum import KNXIPServiceType


class DeviceConfigurationRequest(KNXIPBody):
    """Representation of a KNX Device Configuration Request."""

    SERVICE_TYPE = KNXIPServiceType.DEVICE_CONFIGURATION_REQUEST
    HEADER_LENGTH = 4

    def __init__(
        self,
        communication_channel_id: int = 1,
        sequence_counter: int = 0,
        raw_cemi: bytes = b"",
    ) -> None:
        """Initialize DeviceConfigurationRequest object."""
        self.communication_channel_id = communication_channel_id
        self.sequence_counter = sequence_counter
        self.raw_cemi = raw_cemi

    def calculated_length(self) -> int:
        """Get length of KNX/IP body."""
        return DeviceConfigurationRequest.HEADER_LENGTH + len(self.raw_cemi)

    def from_knx(self, raw: bytes) -> int:
        """Parse/deserialize from KNX/IP raw data."""
        if raw[0] != DeviceConfigurationRequest.HEADER_LENGTH:
            raise CouldNotParseKNXIP("connection header wrong length")
        if len(raw) < DeviceConfigurationRequest.HEADER_LENGTH:
            raise CouldNotParseKNXIP("connection header wrong length")
        self.communication_channel_id = raw[1]
        self.sequence_counter = raw[2]
        # raw[3] is reserved
        self.raw_cemi = raw[DeviceConfigurationRequest.HEADER_LENGTH :]
        return len(raw)

    def to_knx(self) -> bytes:
        """Serialize to KNX/IP raw data."""
        return (
            bytes(
                (
                    DeviceConfigurationRequest.HEADER_LENGTH,
                    self.communication_channel_id,
                    self.sequence_counter,
                    0x00,  # Reserved
                )
            )
            + self.raw_cemi
        )

    def __repr__(self) -> str:
        """Return object as readable string."""
        return (
            "<DeviceConfigurationRequest "
            f'communication_channel_id="{self.communication_channel_id}" '
            f'sequence_counter="{self.sequence_counter}" '
            f'cemi="{self.raw_cemi.hex()}" />'
        )



"""Abstraction to send SessionRequest and wait for SessionResponse."""

from __future__ import annotations

from typing import TYPE_CHECKING

from xknx.knxip import KNXIPFrame, SessionRequest, SessionResponse

from .request_response import RequestResponse

if TYPE_CHECKING:
    from xknx.io.transport import KNXIPTransport


class Session(RequestResponse):
    """Class to send a SessionRequest and wait for SessionResponse."""

    def __init__(
        self,
        transport: KNXIPTransport,
        ecdh_client_public_key: bytes,
    ) -> None:
        """Initialize Session class."""
        # TODO: increase timeout to timeoutAuthentication: 10sec ?
        super().__init__(transport, SessionResponse)
        self.ecdh_client_public_key = ecdh_client_public_key
        # TODO: make RequestResponse generic for response class
        # maybe replace self.success with self.response None check
        # remove on_success_hook in favour of using knxipframe.body directly
        self.response: SessionResponse | None = None

    def create_knxipframe(self) -> KNXIPFrame:
        """Create KNX/IP Frame object to be sent to device."""
        return KNXIPFrame.init_from_body(
            SessionRequest(ecdh_client_public_key=self.ecdh_client_public_key)
        )

    def on_success_hook(self, knxipframe: KNXIPFrame) -> None:
        """Set communication channel and identifier after having received a valid answer."""
        assert isinstance(knxipframe.body, SessionResponse)
        self.response = knxipframe.body



"""Module for managing remote value with payload length based DPT detection."""

from __future__ import annotations

from collections.abc import Iterable
from typing import TYPE_CHECKING

from xknx.dpt import DPTArray, DPTBinary, DPTNumeric
from xknx.exceptions import ConversionError, CouldNotParseTelegram

from .remote_value import GroupAddressesType, RemoteValue, RVCallbackType

if TYPE_CHECKING:
    from xknx.xknx import XKNX


class RemoteValueByLength(RemoteValue[float]):
    """RemoteValue with DPT detection based on payload length of first received value."""

    def __init__(
        self,
        xknx: XKNX,
        dpt_classes: Iterable[type[DPTNumeric]],
        group_address: GroupAddressesType = None,
        group_address_state: GroupAddressesType = None,
        sync_state: bool | int | float | str = True,
        device_name: str | None = None,
        feature_name: str | None = None,
        after_update_cb: RVCallbackType[float] | None = None,
    ) -> None:
        """Initialize RemoteValueByLength class."""
        _payload_lengths = set()
        for dpt_class in dpt_classes:
            if (
                not issubclass(dpt_class, DPTNumeric)
                or dpt_class.payload_type is not DPTArray
            ):
                raise ConversionError(
                    "Only DPTNumeric subclasses with payload_type DPTArray are supported"
                )
            if dpt_class.payload_length in _payload_lengths:
                raise ConversionError(
                    f"Duplicate payload_length {dpt_class.payload_length} in {dpt_classes}"
                )
            _payload_lengths.add(dpt_class.payload_length)

        super().__init__(
            xknx,
            group_address,
            group_address_state,
            sync_state=sync_state,
            device_name=device_name,
            feature_name=feature_name,
            after_update_cb=after_update_cb,
        )

        self._dpt_classes = dpt_classes
        self._internal_dpt_class: type[DPTNumeric] | None = None

    def to_knx(self, value: float) -> DPTArray:
        """Convert value to payload."""
        if self._internal_dpt_class is None:
            raise ConversionError(
                f"RemoteValue DPT not initialized for {self.device_name}"
            )
        return self._internal_dpt_class.to_knx(value)

    def from_knx(self, payload: DPTArray | DPTBinary) -> float:
        """Convert current payload to value."""
        if self._internal_dpt_class is None:
            self._internal_dpt_class = self._determine_dpt_class(payload)

        return self._internal_dpt_class.from_knx(payload)

    @property
    def unit_of_measurement(self) -> str | None:
        """Return the unit of measurement."""
        if not self._internal_dpt_class:
            return None
        return self._internal_dpt_class.unit

    @property
    def ha_device_class(self) -> str | None:
        """Return a string representing the home assistant device class."""
        if not self._internal_dpt_class:
            return None
        return getattr(self._internal_dpt_class, "ha_device_class", None)

    def _determine_dpt_class(self, payload: DPTArray | DPTBinary) -> type[DPTNumeric]:
        """Test if telegram payload may be parsed."""
        if isinstance(payload, DPTArray):
            try:
                return next(
                    dpt_class
                    for dpt_class in self._dpt_classes
                    if dpt_class.payload_type is DPTArray
                    and dpt_class.payload_length == len(payload.value)
                )
            except StopIteration:
                pass

        raise CouldNotParseTelegram("Payload invalid", payload=str(payload))



"""Unit test for task registry."""

import asyncio
from unittest.mock import AsyncMock, Mock

import pytest

from xknx import XKNX
from xknx.core import XknxConnectionState
from xknx.core.task_registry import Task

from ..conftest import EventLoopClockAdvancer


class TestTaskRegistry:
    """Test class for task registry."""

    @pytest.mark.parametrize("target", ["async", "sync"])
    async def test_start_task(self, target: str) -> None:
        """Test start_task."""
        xknx = XKNX()
        mock = AsyncMock() if target == "async" else Mock()

        task = Task(
            name="test",
            target=mock,
        )
        xknx.task_registry.start_task(task)
        assert len(xknx.task_registry.tasks) == 1

        assert not task.done()
        assert mock.call_count == 0
        if target == "async":
            assert mock.await_count == 0

        await xknx.task_registry.block_till_done()
        assert task.done()
        assert mock.call_count == 1
        if target == "async":
            assert mock.await_count == 1

    async def test_remove_task(self) -> None:
        """Test remove_task after starting task."""
        xknx = XKNX()
        mock = Mock()

        task = Task(
            name="test",
            target=mock,
        )
        xknx.task_registry.start_task(task)
        assert len(xknx.task_registry.tasks) == 1
        xknx.task_registry.remove_task(task)
        assert len(xknx.task_registry.tasks) == 0
        assert task.done()

        assert mock.call_count == 0

    #
    # TEST START/STOP
    #
    async def test_stop(self) -> None:
        """Test stop."""
        xknx = XKNX()

        async def callback() -> None:
            """Reset tasks."""
            await asyncio.sleep(1)
            raise AssertionError("Task should have been cancelled")

        task = Task(
            name="test",
            target=callback,
        )
        xknx.task_registry.start_task(task)

        assert len(xknx.task_registry.tasks) == 1
        xknx.task_registry.stop()
        assert len(xknx.task_registry.tasks) == 0
        assert task.done()

    #
    # TEST CONNECTION HANDLING
    #
    async def test_reconnect_handling(
        self, time_travel: EventLoopClockAdvancer
    ) -> None:
        """Test reconnect handling."""

        xknx = XKNX()
        xknx.task_registry.start()
        assert len(xknx.connection_manager._connection_state_changed_cbs) == 1
        xknx.connection_manager.connection_state_changed(XknxConnectionState.CONNECTED)

        test = 0

        async def callback() -> None:
            """Reset tasks."""
            nonlocal test
            try:
                while True:
                    await asyncio.sleep(100)
                    test += 1
            except asyncio.CancelledError:
                test -= 1

        task = Task(
            name="test",
            target=callback,
            restart_after_reconnect=True,
        )
        xknx.task_registry.start_task(task)
        assert len(xknx.task_registry.tasks) == 1
        assert task._task is not None
        await time_travel(100)
        assert test == 1

        xknx.connection_manager.connection_state_changed(
            XknxConnectionState.DISCONNECTED
        )
        await asyncio.sleep(0)  # iterate loop to cancel task
        assert task._task is None
        assert test == 0

        xknx.connection_manager.connection_state_changed(XknxConnectionState.CONNECTED)
        assert task._task is not None
        assert test == 0

        await time_travel(100)
        assert test == 1
        assert len(xknx.task_registry.tasks) == 1

        xknx.task_registry.stop()
        assert len(xknx.task_registry.tasks) == 0
        assert task._task is None
        await asyncio.sleep(0)  # iterate loop to cancel task
        assert test == 0
        assert len(xknx.connection_manager._connection_state_changed_cbs) == 0

    async def test_wait_before_start(self, time_travel: EventLoopClockAdvancer) -> None:
        """Test wait_before_start argument."""
        xknx = XKNX()
        mock = AsyncMock()
        wait_time = 5

        task = Task(
            name="test",
            target=mock,
            wait_before_start=wait_time,
        )
        xknx.task_registry.start_task(task)

        # Task should not be called immediately
        assert mock.call_count == 0
        assert mock.await_count == 0
        assert not task.done()

        # Task should not be called after half the wait time
        await time_travel(wait_time / 2)
        assert mock.call_count == 0
        assert mock.await_count == 0
        assert not task.done()

        # Task should be called after the full wait time
        await time_travel(wait_time / 2)
        assert mock.call_count == 1
        assert mock.await_count == 1
        assert task.done()

        # restart task - verify target can be called again
        # after done and wait_before_start applies again
        task.restart()
        await time_travel(wait_time / 2)
        assert mock.call_count == 1
        await time_travel(wait_time / 2)
        assert mock.call_count == 2
        assert mock.await_count == 2
        assert task.done()

    async def test_wait_for_connection(self) -> None:
        """Test wait_for_connection argument."""
        xknx = XKNX()
        mock = Mock()

        task = Task(
            name="test",
            target=mock,
            wait_for_connection=True,
        )
        xknx.task_registry.start_task(task)

        # Task should not be called immediately when not connected
        await asyncio.sleep(0)
        assert mock.call_count == 0
        assert not task.done()

        # Task should be called after connection is established
        xknx.connection_manager.connection_state_changed(XknxConnectionState.CONNECTED)
        await asyncio.sleep(0)
        assert mock.call_count == 1
        assert task.done()

    async def test_repeat_after(self, time_travel: EventLoopClockAdvancer) -> None:
        """Test repeat_after argument."""
        xknx = XKNX()
        mock = Mock()
        repeat_after = 5

        task = Task(
            name="test",
            target=mock,
            repeat_after=repeat_after,
        )
        xknx.task_registry.start_task(task)

        await asyncio.sleep(0)
        assert mock.call_count == 1
        assert not task.done()

        await time_travel(repeat_after)
        assert mock.call_count == 2
        assert not task.done()

        await time_travel(repeat_after)
        assert mock.call_count == 3
        assert not task.done()

        task.cancel()
        await asyncio.sleep(0)
        assert task.done()

    async def test_wait_before_start_no_connection_with_restart(
        self, time_travel: EventLoopClockAdvancer
    ) -> None:
        """Test restart_after_reconnect when wait_before_start ends without connection."""
        xknx = XKNX()
        xknx.task_registry.start()
        mock = Mock()
        wait_time = 5

        task = Task(
            name="test",
            target=mock,
            restart_after_reconnect=True,
            wait_before_start=wait_time,
            wait_for_connection=True,
        )
        xknx.task_registry.start_task(task)

        await time_travel(wait_time)
        await asyncio.sleep(0)
        assert mock.call_count == 0
        assert task.done()

        # restarted when connection is reestablished - wait_before_start should apply again
        xknx.connection_manager.connection_state_changed(XknxConnectionState.CONNECTED)
        await asyncio.sleep(0)
        assert not task.done()
        assert mock.call_count == 0

        await time_travel(wait_time)
        await asyncio.sleep(0)
        assert mock.call_count == 1
        assert task.done()

    async def test_start_before_register(self) -> None:
        """Test that start() without registration in TaskRegistry raises RuntimeError."""
        task = Task(
            name="test",
            target=Mock(),
        )

        with pytest.raises(RuntimeError, match="Task must be registered before start"):
            task._start()

    async def test_background(self, time_travel: EventLoopClockAdvancer) -> None:
        """Test running background task."""
        test_time = 10

        async def callback() -> None:
            """Do nothing."""
            await asyncio.sleep(test_time)

        xknx = XKNX()
        xknx.task_registry.background(callback())
        assert len(xknx.task_registry._background_task) == 1
        task = next(iter(xknx.task_registry._background_task))
        assert not task.done()

        await time_travel(test_time / 2)
        assert not task.done()
        assert task in xknx.task_registry._background_task

        await time_travel(test_time / 2)
        # after task is finished it should remove itself from the background registry
        assert task.done()
        assert task not in xknx.task_registry._background_task



"""Unit test for KNX DPT 18 objects."""

from typing import Any

import pytest

from xknx.dpt import DPTArray, DPTBinary, DPTSceneControl, SceneControl
from xknx.exceptions import ConversionError, CouldNotParseTelegram


class TestSceneControl:
    """Test SceneControl class."""

    @pytest.mark.parametrize(
        ("data", "value"),
        [
            (
                {"scene_number": 5, "learn": False},
                SceneControl(5, False),
            ),
            (
                {"scene_number": 2},
                SceneControl(scene_number=2),
            ),
            (
                {"scene_number": 17, "learn": True},
                SceneControl(scene_number=17, learn=True),
            ),
        ],
    )
    def test_dict(self, data: dict[str, Any], value: SceneControl) -> None:
        """Test from_dict and as_dict methods."""
        test_value = SceneControl.from_dict(data)
        assert test_value == value
        # learn defaults to `False`
        default_dict = {"learn": False}
        assert value.as_dict() == default_dict | data

    @pytest.mark.parametrize(
        "data",
        [
            # invalid data
            {"learn": False},
            {"scene_number": "a"},
            {"scene_number": None, "learn": True},
            {"scene_number": 1, "learn": "a"},
            {"scene_number": 1, "learn": None},
        ],
    )
    def test_dict_invalid(self, data: dict[str, Any]) -> None:
        """Test from_dict and as_dict methods."""
        with pytest.raises(ValueError):
            SceneControl.from_dict(data)


class TestDPTSceneControl:
    """Test class for KNX DPTSceneControl objects."""

    @pytest.mark.parametrize(
        ("value", "raw"),
        [
            (SceneControl(1, False), (0b00000000,)),
            (SceneControl(1, True), (0b10000000,)),
            (SceneControl(64, False), (0b00111111,)),
            (SceneControl(64, True), (0b10111111,)),
        ],
    )
    def test_parse(self, value: SceneControl, raw: tuple[int]) -> None:
        """Test DPTTariffActiveEnergy parsing and streaming."""
        knx_value = DPTSceneControl.to_knx(value)
        assert knx_value == DPTArray(raw)
        assert DPTSceneControl.from_knx(knx_value) == value

    @pytest.mark.parametrize(
        ("value", "raw"),
        [
            ({"scene_number": 17, "learn": False}, (0x10,)),
            ({"scene_number": 21, "learn": True}, (0x94,)),
        ],
    )
    def test_to_knx_from_dict(self, value: dict[str, Any], raw: tuple[int]) -> None:
        """Test DPTTariffActiveEnergy parsing from a dict."""
        assert DPTSceneControl.to_knx(value) == DPTArray(raw)

    @pytest.mark.parametrize(
        ("value"),
        [
            SceneControl(scene_number=0, learn=False),
            SceneControl(scene_number=65, learn=True),
        ],
    )
    def test_to_knx_limits(self, value: SceneControl) -> None:
        """Test initialization of DPTTariffActiveEnergy with wrong value."""
        with pytest.raises(ConversionError):
            DPTSceneControl.to_knx(value)

    def test_from_knx_wrong_value(self) -> None:
        """Test DPTTariffActiveEnergy parsing with wrong value."""
        with pytest.raises(CouldNotParseTelegram):
            DPTSceneControl.from_knx(DPTArray((0xFF, 0x4E)))
        with pytest.raises(CouldNotParseTelegram):
            DPTSceneControl.from_knx(DPTBinary(True))



"""Unit test for xknx.io.connection module."""

from xknx.io.connection import ConnectionConfig, SecureConfig


def test_connection_config_compare() -> None:
    """Test ConnectionConfig comparison."""
    assert ConnectionConfig() == ConnectionConfig()
    assert ConnectionConfig() != ConnectionConfig(individual_address="1.1.1")


def test_secure_config_compare() -> None:
    """Test SecureConfig comparison."""
    assert SecureConfig() == SecureConfig()
    assert SecureConfig() != SecureConfig(user_id=5)



"""Unit test for Telegram objects."""

from xknx.dpt import DPTBinary, DPTSwitch
from xknx.telegram import (
    GroupAddress,
    IndividualAddress,
    Telegram,
    TelegramDecodedData,
    TelegramDirection,
)
from xknx.telegram.apci import GroupValueRead, GroupValueWrite
from xknx.telegram.tpci import TConnect, TDisconnect


class TestTelegram:
    """Test class for Telegram objects."""

    #
    # EQUALITY
    #
    def test_telegram_equal(self) -> None:
        """Test equals operator."""
        test = Telegram(GroupAddress("1/2/3"), payload=GroupValueRead())
        telegram_1 = Telegram(GroupAddress("1/2/3"), payload=GroupValueRead())
        assert test == telegram_1
        # decoded_data and data_secure should not be considered for equality (although
        # decoded_data doesn't make sense for a GroupValueRead, this is just for testing the equality operator)
        telegram_1.decoded_data = TelegramDecodedData(transcoder=DPTSwitch, value=False)
        telegram_1.data_secure = True
        assert test == telegram_1

    def test_telegram_not_equal(self) -> None:
        """Test not equals operator."""
        assert Telegram(GroupAddress("1/2/3"), payload=GroupValueRead()) != Telegram(
            GroupAddress("1/2/4"), payload=GroupValueRead()
        )
        assert Telegram(GroupAddress("1/2/3"), payload=GroupValueRead()) != Telegram(
            GroupAddress("1/2/3"), payload=GroupValueWrite(DPTBinary(1))
        )
        assert Telegram(GroupAddress("1/2/3"), payload=GroupValueRead()) != Telegram(
            GroupAddress("1/2/3"),
            TelegramDirection.INCOMING,
            payload=GroupValueRead(),
        )
        assert Telegram(IndividualAddress(1), tpci=TConnect()) != Telegram(
            IndividualAddress(1), tpci=TDisconnect()
        )



"""Example for Weather device. See docs/weather.md for a detailed explanation."""

import asyncio
import logging

from xknx import XKNX
from xknx.devices import Weather

logging.basicConfig(level=logging.WARNING)


async def main() -> None:
    """Connect to KNX/IP device and create a weather device and read its sensors."""
    xknx = XKNX()
    await xknx.start()

    weather = Weather(
        xknx,
        "Home",
        group_address_temperature="7/0/1",
        group_address_brightness_south="7/0/5",
        group_address_brightness_east="7/0/4",
        group_address_brightness_west="7/0/3",
        group_address_wind_speed="7/0/2",
        group_address_wind_bearing="7/0/6",
        group_address_day_night="7/0/7",
        group_address_rain_alarm="7/0/0",
    )
    xknx.devices.async_add(weather)

    await weather.sync(wait_for_result=True)
    print(weather.max_brightness)
    print(weather.ha_current_state())
    print(weather)

    await xknx.stop()


asyncio.run(main())



"""Module for KNX/IP Enum classes."""

from enum import Enum


class KNXIPServiceType(Enum):
    """Enum class for KNX/IP Service Types."""

    # 0x02 Core services
    SEARCH_REQUEST = 0x0201
    SEARCH_RESPONSE = 0x0202
    DESCRIPTION_REQUEST = 0x0203
    DESCRIPTION_RESPONSE = 0x0204
    CONNECT_REQUEST = 0x0205
    CONNECT_RESPONSE = 0x0206
    CONNECTIONSTATE_REQUEST = 0x0207
    CONNECTIONSTATE_RESPONSE = 0x0208
    DISCONNECT_REQUEST = 0x0209
    DISCONNECT_RESPONSE = 0x020A
    SEARCH_REQUEST_EXTENDED = 0x020B
    SEARCH_RESPONSE_EXTENDED = 0x020C
    # 0x03 Device Management services
    DEVICE_CONFIGURATION_REQUEST = 0x0310
    DEVICE_CONFIGURATION_ACK = 0x0311
    # 0x04 Tunneling services
    TUNNELLING_REQUEST = 0x0420
    TUNNELLING_ACK = 0x0421
    TUNNELLING_FEATURE_GET = 0x0422
    TUNNELLING_FEATURE_RESPONSE = 0x0423
    TUNNELLING_FEATURE_SET = 0x0424
    TUNNELLING_FEATURE_INFO = 0x0425
    # 0x05 Routing services
    ROUTING_INDICATION = 0x0530
    ROUTING_LOST_MESSAGE = 0x0531
    ROUTING_BUSY = 0x0532
    ROUTING_SYSTEM_BROADCAST = 0x0533
    # 0x06 Remote Logging services
    # 0x07 Remote Configuration and Diagnosis services
    REMOTE_DIAG_REQUEST = 0x0740
    REMOTE_DIAG_RESPONSE = 0x0741
    REMOTE_CONFIG_REQUEST = 0x0742
    REMOTE_RESET_REQUEST = 0x0743
    # 0x08 Object Server services
    # 0x09 Security services
    SECURE_WRAPPER = 0x0950
    SESSION_REQUEST = 0x0951
    SESSION_RESPONSE = 0x0952
    SESSION_AUTHENTICATE = 0x0953
    SESSION_STATUS = 0x0954
    TIMER_NOTIFY = 0x0955


class ConnectRequestType(Enum):
    """Enum class for KNX/IP Connect Request Types."""

    # Data connection used to configure a KNXnet/IP device
    DEVICE_MGMT_CONNECTION = 0x03
    # Data connection used to forward KNX telegrams between
    # two KNXnet/IP devices.
    TUNNEL_CONNECTION = 0x04
    # Data connection used for configuration and data transfer
    # with a remote logging server.
    REMLOG_CONNECTION = 0x06
    # Data connection used for data transfer with a remote
    # configuration server.
    REMCONF_CONNECTION = 0x07
    # Data connection used for configuration and data transfer
    # with an Object Server in a KNXnet/IP device.
    OBJSVR_CONNECTION = 0x08


class TunnellingLayer(Enum):
    """Enum class for KNX/IP Tunnelling Layer."""

    # Data Link Layer tunnel
    DATA_LINK_LAYER = 0x02
    # Raw tunnel
    RAW_LAYER = 0x04
    # Busmonitor tunnel
    BUSMONITOR_LAYER = 0x80


class DIBTypeCode(Enum):
    """Enum class for KNX/IP DIB Type Codes."""

    # Device information e.g. KNX medium.
    DEVICE_INFO = 0x01
    # Service families supported by the device.
    SUPP_SVC_FAMILIES = 0x02
    # IP configuration
    IP_CONFIG = 0x03
    # current configuration
    IP_CUR_CONFIG = 0x04
    # KNX addresses
    KNX_ADDRESSES = 0x05
    # KNX IP Secure
    SECURED_SERVICE_FAMILIES = 0x06
    TUNNELING_INFO = 0x07
    ADDITIONAL_DEVICE_INFO = 0x08
    # DIB structure for further data defined by device manufacturer.
    MFR_DATA = 0xFE


class HostProtocol(Enum):
    """Enum class for host protocol."""

    IPV4_UDP = 0x01
    IPV4_TCP = 0x02


class KNXMedium(Enum):
    """Enum class for KNX Medium."""

    TP1 = 0x02
    PL110 = 0x04
    RF = 0x10
    KNX_IP = 0x20


class DIBServiceFamily(Enum):
    """Enum class for KNX/IP DIB Service Family."""

    #  Core
    CORE = 0x02
    # Device Management
    DEVICE_MANAGEMENT = 0x03
    # Tunnelling
    TUNNELING = 0x04
    # Routing
    ROUTING = 0x05
    # Remote Logging
    REMOTE_LOGGING = 0x06
    # Configuration and Diagnosis
    REMOTE_CONFIGURATION_DIAGNOSIS = 0x07
    # Object Server'.
    OBJECT_SERVER = 0x08
    # Security - Extended search response only
    SECURITY = 0x09


class SecureSessionStatusCode(Enum):
    """Enum class for KNX/IP Secure session status codes."""

    # The user could successfully be authenticated
    STATUS_AUTHENTICATION_SUCCESS = 0x00
    # An error occurred during secure session handshake
    STATUS_AUTHENTICATION_FAILED = 0x01
    # The session is not (yet) authenticated
    STATUS_UNAUTHENTICATED = 0x02
    # A timeout occurred during secure session handshake
    STATUS_TIMEOUT = 0x03
    # Prevent inactivity on the secure session closing it with timeout error
    STATUS_KEEPALIVE = 0x04
    # The secure session shall be closed
    STATUS_CLOSE = 0x05


class SearchRequestParameterType(Enum):
    """Search Request Parameter (SRP) is used to transfer additional information in a KNXnet/IP SEARCH_REQUEST_EXTENDED."""

    # Used to test behavior of the KNXnet/IP server for unknown SRPs. Don't use!
    INVALID = 0x00
    # Select only KNXnet/IP Servers that are currently in programming mode
    SELECT_BY_PROGRAMMING_MODE = 0x01
    # Select only KNXnet/IP Servers that have the given MAC address
    SELECT_BY_MAC_ADDRESS = 0x02
    # Select only KNXnet/IP Servers that support the given DIBServiceFamily in a given version
    SELECT_BY_SERVICE = 0x03
    # The Client shall include this SRP to indicate that it is interested in the listed DIBs
    REQUEST_DIBS = 0x04


class TunnellingFeatureType(Enum):
    """Enum class for KNX/IP Tunnel Features services."""

    # Getting the supported EMI type(s) (only cEMI allowed)
    SUPPORTED_EMI_TYPE = 0x01
    # Getting the local Device Descriptor Type 0 for possible local device management
    DEVICE_DESCRIPTOR_TYPE_0 = 0x02
    # Getting and informing on the bus connection status
    BUS_CONNECTION_STATUS = 0x03
    # Getting the manufacturer code of the Bus Access Server
    MANUFACTURER_CODE = 0x04
    # Getting and Setting the EMI type to use
    ACTIVE_EMI_TYPE = 0x05
    # Getting, Setting and informing on the Interface Individual Address
    INTERFACE_INDIVIDUAL_ADDRESS = 0x06
    # Getting and informing on the maximum APDU length
    MAX_APDU_LENGTH = 0x07
    # Getting and Setting the Interface Feature Info service Enable
    INTERFACE_FEATURE_INFO_ENABLE = 0x08



"""Module for broadcasting date/time to the KNX bus, optionally broadcasting localtime periodically."""

from __future__ import annotations

from abc import abstractmethod
from collections.abc import Iterator
import datetime
import logging
import time
from typing import TYPE_CHECKING, Any, Generic, TypeVar

from xknx.core import Task
from xknx.dpt.dpt_10 import KNXDay, KNXTime
from xknx.dpt.dpt_11 import KNXDate
from xknx.dpt.dpt_19 import KNXDateTime, KNXDayOfWeek
from xknx.remote_value import (
    GroupAddressesType,
    RemoteValue,
    RemoteValueDate,
    RemoteValueDateTime,
    RemoteValueTime,
)
from xknx.typing import Self

from .device import Device, DeviceCallbackType

if TYPE_CHECKING:
    from xknx.telegram import Telegram
    from xknx.xknx import XKNX

logger = logging.getLogger("xknx.log")


BROADCAST_MINUTES = 60

_RemoteValueTimeT = TypeVar(
    "_RemoteValueTimeT", RemoteValueTime, RemoteValueDate, RemoteValueDateTime
)


class _DateTimeBase(Device, Generic[_RemoteValueTimeT]):
    """Base class for virtual date/time device."""

    _remote_value_cls: type[_RemoteValueTimeT]  # set in subclass

    remote_value: _RemoteValueTimeT

    def __init__(
        self,
        xknx: XKNX,
        name: str,
        localtime: bool | datetime.tzinfo = True,
        group_address: GroupAddressesType = None,
        group_address_state: GroupAddressesType = None,
        respond_to_read: bool = False,
        sync_state: bool | int | float | str = True,
        device_updated_cb: DeviceCallbackType[Self] | None = None,
    ) -> None:
        """Initialize DateTime class."""
        super().__init__(xknx, name, device_updated_cb)
        self.localtime = bool(localtime)
        self._localtime_zone = (
            localtime if isinstance(localtime, datetime.tzinfo) else None
        )
        if localtime and bool(group_address_state):
            logger.warning(
                "State address invalid in %s device when using `localtime=True`. Ignoring `group_address_state=%s` argument.",
                self.__class__.__name__,
                group_address_state,
            )
            # state address invalid for localtime - therefore sync_state doesn't apply for localtime
            group_address_state = None
        self.respond_to_read = respond_to_read
        self.remote_value = self._remote_value_cls(
            xknx,
            group_address=group_address,
            group_address_state=group_address_state,
            sync_state=sync_state,
            device_name=name,
            after_update_cb=self.after_update,
        )
        self._broadcast_task: Task | None = None

    def _iter_remote_values(self) -> Iterator[RemoteValue[Any]]:
        """Iterate the devices RemoteValue classes."""
        yield self.remote_value

    @property
    @abstractmethod
    def value(self) -> datetime.datetime | datetime.date | datetime.time | None:
        """Return the current date/time value."""

    def async_start_tasks(self) -> None:
        """Create an asyncio.Task for broadcasting local time periodically if `localtime` is set."""
        if not self.localtime:
            return None

        self._broadcast_task = Task(
            name=f"datetime.broadcast_{id(self)}",
            target=self.broadcast_localtime,
            restart_after_reconnect=True,
            wait_for_connection=True,
            repeat_after=BROADCAST_MINUTES * 60,
        )
        self.xknx.task_registry.start_task(self._broadcast_task)

    def async_remove_tasks(self) -> None:
        """Stop background tasks of device."""
        if self._broadcast_task is not None:
            self.xknx.task_registry.remove_task(self._broadcast_task)
            self._broadcast_task = None

    @abstractmethod
    def broadcast_localtime(self, response: bool = False) -> None:
        """Broadcast the local time to KNX bus."""
        # self.remote_value.set(now, response=response)

    @abstractmethod
    async def set(self, value: Any) -> None:
        """Set time and send to KNX bus."""
        # self.remote_value.set(value)

    def process_group_write(self, telegram: Telegram) -> None:
        """Process incoming and outgoing GROUP WRITE telegram."""
        self.remote_value.process(telegram)

    def process_group_read(self, telegram: Telegram) -> None:
        """Process incoming GROUP READ telegram."""
        if self.localtime:
            self.broadcast_localtime(True)
        elif (
            self.respond_to_read
            and telegram.destination_address == self.remote_value.group_address
        ):
            self.remote_value.respond()

    async def sync(self, wait_for_result: bool = False) -> None:
        """Read state of device from KNX bus. Used here to broadcast time to KNX bus."""
        if self.localtime:
            self.broadcast_localtime(response=False)
        else:
            await super().sync(wait_for_result)

    def __str__(self) -> str:
        """Return object as readable string."""
        return (
            f'<{self.__class__.__name__} name="{self.name}" '
            f"remote_value={self.remote_value.group_addr_str()} />"
        )


class TimeDevice(_DateTimeBase[RemoteValueTime]):
    """Class for virtual time device."""

    _remote_value_cls = RemoteValueTime

    @property
    def value(self) -> datetime.time | None:
        """Return the current time value."""
        return self.remote_value.value.as_time() if self.remote_value.value else None

    async def set(self, value: KNXTime | datetime.time) -> None:
        """Set time and send to KNX bus."""
        if isinstance(value, datetime.time):
            value = KNXTime.from_time(value)
        self.remote_value.set(value)

    def broadcast_localtime(self, response: bool = False) -> None:
        """Broadcast the local time to KNX bus."""
        now = datetime.datetime.now(self._localtime_zone)
        knx_time = KNXTime.from_time(now.time())
        knx_time.day = KNXDay(now.weekday() + 1)
        self.remote_value.set(knx_time, response=response)


class DateDevice(_DateTimeBase[RemoteValueDate]):
    """Class for virtual date device."""

    _remote_value_cls = RemoteValueDate

    @property
    def value(self) -> datetime.date | None:
        """Return the current time value."""
        return self.remote_value.value.as_date() if self.remote_value.value else None

    async def set(self, value: KNXDate | datetime.date) -> None:
        """Set date and send to KNX bus."""
        if isinstance(value, datetime.date):
            value = KNXDate.from_date(value)
        self.remote_value.set(value)

    def broadcast_localtime(self, response: bool = False) -> None:
        """Broadcast the local date to KNX bus."""
        now = datetime.datetime.now(self._localtime_zone)
        self.remote_value.set(KNXDate.from_date(now.date()), response=response)


class DateTimeDevice(_DateTimeBase[RemoteValueDateTime]):
    """Class for virtual date/time device."""

    _remote_value_cls = RemoteValueDateTime

    timezone: datetime.tzinfo | None = None

    @property
    def value(self) -> datetime.datetime | None:
        """Return the current time value."""
        return (
            self.remote_value.value.as_datetime() if self.remote_value.value else None
        )

    async def set(self, value: KNXDateTime | datetime.datetime) -> None:
        """Set date/time and send to KNX bus."""
        if isinstance(value, datetime.datetime):
            value = KNXDateTime.from_datetime(value)
        self.remote_value.set(value)

    def broadcast_localtime(self, response: bool = False) -> None:
        """Broadcast the local date/time to KNX bus."""
        now = datetime.datetime.now(self._localtime_zone)
        knx_datetime = KNXDateTime.from_datetime(now)
        knx_datetime.day_of_week = KNXDayOfWeek(now.weekday() + 1)
        if self._localtime_zone is not None:
            dst = self._localtime_zone.dst(now)
            knx_datetime.dst = dst > datetime.timedelta(0) if dst is not None else False
        else:
            time_now = time.localtime()
            knx_datetime.dst = time_now.tm_isdst > 0
        knx_datetime.external_sync = True
        knx_datetime.source_reliable = True
        self.remote_value.set(knx_datetime, response=response)



"""Unit test for RemoteValueScaling objects."""

from xknx import XKNX
from xknx.remote_value import RemoteValueScaling


class TestRemoteValueScaling:
    """Test class for RemoteValueScaling objects."""

    def test_calc_0_10(self) -> None:
        """Test if from/to calculations work with small range."""
        assert RemoteValueScaling._calc_to_knx(0, 10, 0) == 0
        assert RemoteValueScaling._calc_to_knx(0, 10, 1) == 26
        assert RemoteValueScaling._calc_to_knx(0, 10, 9) == 230
        assert RemoteValueScaling._calc_to_knx(0, 10, 10) == 255

        assert RemoteValueScaling._calc_from_knx(0, 10, 0) == 0
        assert RemoteValueScaling._calc_from_knx(0, 10, 1) == 0
        assert RemoteValueScaling._calc_from_knx(0, 10, 12) == 0
        assert RemoteValueScaling._calc_from_knx(0, 10, 13) == 1
        assert RemoteValueScaling._calc_from_knx(0, 10, 254) == 10
        assert RemoteValueScaling._calc_from_knx(0, 10, 255) == 10

    def test_calc_0_100(self) -> None:
        """Test if from/to calculations work range 0-100 with many test cases."""
        assert RemoteValueScaling._calc_to_knx(0, 100, 0) == 0
        assert RemoteValueScaling._calc_to_knx(0, 100, 1) == 3
        assert RemoteValueScaling._calc_to_knx(0, 100, 2) == 5
        assert RemoteValueScaling._calc_to_knx(0, 100, 3) == 8
        assert RemoteValueScaling._calc_to_knx(0, 100, 30) == 76
        assert RemoteValueScaling._calc_to_knx(0, 100, 50) == 128
        assert RemoteValueScaling._calc_to_knx(0, 100, 70) == 178
        assert RemoteValueScaling._calc_to_knx(0, 100, 97) == 247
        assert RemoteValueScaling._calc_to_knx(0, 100, 98) == 250
        assert RemoteValueScaling._calc_to_knx(0, 100, 99) == 252
        assert RemoteValueScaling._calc_to_knx(0, 100, 100) == 255

        assert RemoteValueScaling._calc_from_knx(0, 100, 0) == 0
        assert RemoteValueScaling._calc_from_knx(0, 100, 1) == 0
        assert RemoteValueScaling._calc_from_knx(0, 100, 2) == 1
        assert RemoteValueScaling._calc_from_knx(0, 100, 3) == 1
        assert RemoteValueScaling._calc_from_knx(0, 100, 4) == 2
        assert RemoteValueScaling._calc_from_knx(0, 100, 5) == 2
        assert RemoteValueScaling._calc_from_knx(0, 100, 76) == 30
        assert RemoteValueScaling._calc_from_knx(0, 100, 128) == 50
        assert RemoteValueScaling._calc_from_knx(0, 100, 178) == 70
        assert RemoteValueScaling._calc_from_knx(0, 100, 251) == 98
        assert RemoteValueScaling._calc_from_knx(0, 100, 252) == 99
        assert RemoteValueScaling._calc_from_knx(0, 100, 253) == 99
        assert RemoteValueScaling._calc_from_knx(0, 100, 254) == 100
        assert RemoteValueScaling._calc_from_knx(0, 100, 255) == 100

    def test_calc_0_1000(self) -> None:
        """Test if from/to calculations work with large range."""
        assert RemoteValueScaling._calc_to_knx(0, 1000, 0) == 0
        assert RemoteValueScaling._calc_to_knx(0, 1000, 1) == 0
        assert RemoteValueScaling._calc_to_knx(0, 1000, 2) == 1
        assert RemoteValueScaling._calc_to_knx(0, 1000, 3) == 1
        assert RemoteValueScaling._calc_to_knx(0, 1000, 500) == 128
        assert RemoteValueScaling._calc_to_knx(0, 1000, 997) == 254
        assert RemoteValueScaling._calc_to_knx(0, 1000, 998) == 254
        assert RemoteValueScaling._calc_to_knx(0, 1000, 999) == 255
        assert RemoteValueScaling._calc_to_knx(0, 1000, 1000) == 255

        assert RemoteValueScaling._calc_from_knx(0, 1000, 0) == 0
        assert RemoteValueScaling._calc_from_knx(0, 1000, 1) == 4
        assert RemoteValueScaling._calc_from_knx(0, 1000, 2) == 8
        assert RemoteValueScaling._calc_from_knx(0, 1000, 128) == 502
        assert RemoteValueScaling._calc_from_knx(0, 1000, 251) == 984
        assert RemoteValueScaling._calc_from_knx(0, 1000, 252) == 988
        assert RemoteValueScaling._calc_from_knx(0, 1000, 253) == 992
        assert RemoteValueScaling._calc_from_knx(0, 1000, 254) == 996
        assert RemoteValueScaling._calc_from_knx(0, 1000, 255) == 1000

    def test_calc_100_0(self) -> None:
        """Test if from/to calculations work with negative range."""
        assert RemoteValueScaling._calc_to_knx(100, 0, 0) == 255
        assert RemoteValueScaling._calc_to_knx(100, 0, 1) == 252
        assert RemoteValueScaling._calc_to_knx(100, 0, 2) == 250
        assert RemoteValueScaling._calc_to_knx(100, 0, 3) == 247
        assert RemoteValueScaling._calc_to_knx(100, 0, 30) == 178
        assert RemoteValueScaling._calc_to_knx(100, 0, 50) == 128
        assert RemoteValueScaling._calc_to_knx(100, 0, 70) == 76
        assert RemoteValueScaling._calc_to_knx(100, 0, 97) == 8
        assert RemoteValueScaling._calc_to_knx(100, 0, 98) == 5
        assert RemoteValueScaling._calc_to_knx(100, 0, 99) == 3
        assert RemoteValueScaling._calc_to_knx(100, 0, 100) == 0

        assert RemoteValueScaling._calc_from_knx(100, 0, 0) == 100
        assert RemoteValueScaling._calc_from_knx(100, 0, 1) == 100
        assert RemoteValueScaling._calc_from_knx(100, 0, 2) == 99
        assert RemoteValueScaling._calc_from_knx(100, 0, 3) == 99
        assert RemoteValueScaling._calc_from_knx(100, 0, 4) == 98
        assert RemoteValueScaling._calc_from_knx(100, 0, 5) == 98
        assert RemoteValueScaling._calc_from_knx(100, 0, 76) == 70
        assert RemoteValueScaling._calc_from_knx(100, 0, 128) == 50
        assert RemoteValueScaling._calc_from_knx(100, 0, 178) == 30
        assert RemoteValueScaling._calc_from_knx(100, 0, 251) == 2
        assert RemoteValueScaling._calc_from_knx(100, 0, 252) == 1
        assert RemoteValueScaling._calc_from_knx(100, 0, 253) == 1
        assert RemoteValueScaling._calc_from_knx(100, 0, 254) == 0
        assert RemoteValueScaling._calc_from_knx(100, 0, 255) == 0

    def test_calc_100_200(self) -> None:
        """Test if from/to calculations work with range not starting at zero."""
        assert RemoteValueScaling._calc_to_knx(100, 200, 100) == 0
        assert RemoteValueScaling._calc_to_knx(100, 200, 130) == 76
        assert RemoteValueScaling._calc_to_knx(100, 200, 150) == 128
        assert RemoteValueScaling._calc_to_knx(100, 200, 170) == 178
        assert RemoteValueScaling._calc_to_knx(100, 200, 200) == 255

        assert RemoteValueScaling._calc_from_knx(100, 200, 0) == 100
        assert RemoteValueScaling._calc_from_knx(100, 200, 76) == 130
        assert RemoteValueScaling._calc_from_knx(100, 200, 128) == 150
        assert RemoteValueScaling._calc_from_knx(100, 200, 178) == 170
        assert RemoteValueScaling._calc_from_knx(100, 200, 255) == 200

    def test_value_unit(self) -> None:
        """Test for the unit_of_measurement."""
        xknx = XKNX()
        remote_value = RemoteValueScaling(xknx)
        assert remote_value.unit_of_measurement == "%"



"""
Module for managing a remote value typically used within a sensor.

The module maps a given value_type to a DPT class and uses this class
for serialization and deserialization of the KNX value.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from xknx.dpt.dpt import DPTArray, DPTBinary
from xknx.exceptions import ConversionError, CouldNotParseTelegram

from .remote_value import GroupAddressesType, RemoteValue, RVCallbackType

if TYPE_CHECKING:
    from xknx.xknx import XKNX


class RemoteValueRaw(RemoteValue[int]):
    """Abstraction for raw values."""

    def __init__(
        self,
        xknx: XKNX,
        payload_length: int,
        group_address: GroupAddressesType = None,
        group_address_state: GroupAddressesType = None,
        sync_state: bool | int | float | str = True,
        device_name: str | None = None,
        feature_name: str = "Raw",
        after_update_cb: RVCallbackType[int] | None = None,
    ) -> None:
        """Initialize RemoteValueRaw class."""
        self.payload_length = payload_length
        super().__init__(
            xknx,
            group_address,
            group_address_state,
            sync_state=sync_state,
            device_name=device_name,
            feature_name=feature_name,
            after_update_cb=after_update_cb,
        )

    def to_knx(self, value: int) -> DPTArray | DPTBinary:
        """Convert value to payload."""
        if self.payload_length == 0:
            try:
                return DPTBinary(value)
            except TypeError as err:
                raise ConversionError(
                    "Could not init DPTBinary", value=str(value)
                ) from err
        try:
            return DPTArray(value.to_bytes(length=self.payload_length, byteorder="big"))
        except (AttributeError, OverflowError) as err:
            raise ConversionError("Could not init DPTArray", value=str(value)) from err

    def from_knx(self, payload: DPTArray | DPTBinary) -> int:
        """Convert current payload to value."""
        if isinstance(payload, DPTBinary) and self.payload_length == 0:
            return payload.value
        if isinstance(payload, DPTArray) and len(payload.value) == self.payload_length:
            try:
                return int.from_bytes(payload.value, byteorder="big")
            except ValueError as err:
                raise ConversionError(
                    "Could not parse payload", payload=payload
                ) from err
        raise CouldNotParseTelegram("Payload invalid", payload=str(payload))



"""
Module for managing an RGB remote value.

DPT 232.600.
"""

from __future__ import annotations

from xknx.dpt import DPTColorRGB, RGBColor

from .remote_value import RemoteValue


class RemoteValueColorRGB(RemoteValue[RGBColor]):
    """Abstraction for remote value of KNX DPT 232.600 (DPT_Color_RGB)."""

    dpt_class = DPTColorRGB



"""Package for convenience functions for KNX group communication."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from xknx.core.value_reader import ValueReader
from xknx.dpt import DPTArray, DPTBase, DPTBinary
from xknx.telegram import Telegram
from xknx.telegram.address import DeviceAddressableType, parse_device_group_address
from xknx.telegram.apci import GroupValueRead, GroupValueResponse, GroupValueWrite
from xknx.typing import DPTParsable

if TYPE_CHECKING:
    from xknx.xknx import XKNX
logger = logging.getLogger("xknx.tools")


def group_value_read(
    xknx: XKNX,
    group_address: DeviceAddressableType,
) -> None:
    """Send a GroupValueRead telegram."""
    telegram = Telegram(
        destination_address=parse_device_group_address(group_address),
        payload=GroupValueRead(),
    )

    logger.debug("Sending GroupValueRead telegram to %s", group_address)
    xknx.telegrams.put_nowait(telegram)


def group_value_response(
    xknx: XKNX,
    group_address: DeviceAddressableType,
    value: Any,
    value_type: DPTParsable | type[DPTBase] | None = None,
) -> None:
    """Send a GroupValueResponse telegram."""
    payload = _parse_payload(value, value_type)
    telegram = Telegram(
        destination_address=parse_device_group_address(group_address),
        payload=GroupValueResponse(payload),
    )
    logger.debug(
        "Sending GroupValueResponse telegram with value '%s' of type '%s' to %s.",
        value,
        value_type or "raw",
        group_address,
    )
    xknx.telegrams.put_nowait(telegram)


def group_value_write(
    xknx: XKNX,
    group_address: DeviceAddressableType,
    value: Any,
    value_type: DPTParsable | type[DPTBase] | None = None,
) -> None:
    """Send a GroupValueWrite telegram."""
    payload = _parse_payload(value, value_type)
    telegram = Telegram(
        destination_address=parse_device_group_address(group_address),
        payload=GroupValueWrite(payload),
    )
    logger.debug(
        "Sending GroupValueWrite telegram with value '%s' of type '%s' to %s.",
        value,
        value_type or "raw",
        group_address,
    )
    xknx.telegrams.put_nowait(telegram)


async def read_group_value(
    xknx: XKNX,
    group_address: DeviceAddressableType,
    value_type: DPTParsable | type[DPTBase] | None = None,
) -> int | tuple[int, ...] | Any | None:
    """Read a value from a KNX group address."""
    transcoder = _parse_dpt(value_type)
    value_reader = ValueReader(xknx, parse_device_group_address(group_address))
    response = await value_reader.read()
    if response is not None:
        assert isinstance(response.payload, GroupValueWrite | GroupValueResponse)
        if transcoder is not None:
            return transcoder.from_knx(response.payload.value)
        return response.payload.value.value
    return None


def _parse_dpt(value_type: DPTParsable | type[DPTBase] | None) -> type[DPTBase] | None:
    if value_type is None:
        return None
    return DPTBase.get_dpt(value_type)


def _parse_payload(
    value: Any,
    value_type: DPTParsable | type[DPTBase] | None = None,
) -> DPTBinary | DPTArray:
    if isinstance(value, DPTArray | DPTBinary):
        return value
    if transcoder := _parse_dpt(value_type):
        return transcoder.to_knx(value)
    if isinstance(value, int):
        return DPTBinary(value)
    return DPTArray(value)
